#include "Volume.h"
#include <algorithm>
#include <cstring>

constexpr int VolumeStat::histogramSize;

Volume::Volume(size_t width, size_t height, size_t depth,
	vec3 _origin, 
	vec3 _direction0, vec3 _direction1,
	vec3 _voxelSize,
	float intercept /* = 0.0f */, float slope /* = 1.0f */, 
	bool allocateDevice /* = true */)
	: width(width), height(height), depth(depth), nVoxels(width * height * depth),
	origin(_origin), direction0(_direction0), direction1(_direction1), 
	voxelSize(_voxelSize), sliceOrigins(depth), intercept(intercept), slope(slope)
{
	hostPtr = new float[nVoxels];
//	memset(hostPtr, 0, nVoxels * sizeof(float));
	extent.x = voxelSize.x * width;
	extent.y = voxelSize.y * height;
	extent.z = voxelSize.z * depth;
}

Volume::Volume(const Volume& other)
	: width(other.width), height(other.height), depth(other.depth), nVoxels(width* height* depth), 
	origin(other.origin), direction0(other.direction0), direction1(other.direction1), 
	voxelSize(other.voxelSize), sliceOrigins(other.sliceOrigins), extent(other.extent),
	intercept(other.intercept), slope(other.slope), isFlipped(other.isFlipped), axis(other.axis)
{
	hostPtr = new float[nVoxels];
	memcpy(hostPtr, other.hostPtr, nVoxels * sizeof(float));
}

Volume& Volume::operator=(const Volume& other)
{
	if (this != &other)
	{
		delete[] hostPtr;
		width = other.width;
		height = other.height;
		depth = other.depth;
		nVoxels = other.nVoxels;
		origin = other.origin;
		direction0 = other.direction0;
		direction1 = other.direction1;
		voxelSize = other.voxelSize;
		sliceOrigins = other.sliceOrigins;
		extent = other.extent;
		intercept = other.intercept;
		slope = other.slope;
		isFlipped = other.isFlipped;
		axis = other.axis;
		hostPtr = new float[nVoxels];
		memcpy(hostPtr, other.hostPtr, nVoxels * sizeof(float));
	}
	return *this;
}

Volume::~Volume()
{
	delete[] hostPtr;
}

size_t Volume::getWidth() const { return width; }
size_t Volume::getHeight() const { return height; }
size_t Volume::getDepth() const { return depth; }
size_t Volume::getNumberOfVoxels() const { return nVoxels; }
vec3 Volume::getOrigin() const { return origin; }
vec3 Volume::getVoxelSize() const { return voxelSize; }
vec3 Volume::getExtent() const { return extent; }
float Volume::getIntercept() const { return intercept; }
float Volume::getSlope() const { return slope; }
float* Volume::getHostPtr() const { return hostPtr; }

void Volume::print(std::string s) const {
#ifdef PRINT
	std::cout << '\n' << s
		      << getWidth() << " x " << getHeight() << " x " << getDepth()
		      << "\nOrigin: (" << getOrigin().x << ", " << getOrigin().y << ", " << getOrigin().z << "), "
		      << "\nVoxSize: (" << getVoxelSize().x << ", " << getVoxelSize().y << ", " << getVoxelSize().z << "), "
		      << "\nExtent: (" << getExtent().x << ", " << getExtent().y << ", " << getExtent().z << ")\n";
#endif
}

float Volume::getSumOfActivity() const {
	float sum = 0.0f;
	for (size_t i = 0; i < nVoxels; ++i)
	{
		sum += hostPtr[i];
	}
	return sum;
}

float Volume::getAverageActivity() const {
	return getSumOfActivity() / nVoxels;
}

Volume* Volume::getTrueActivities() const 
{
	Volume* trueActivities = new Volume(*this);
	float* data = trueActivities->getHostPtr();
	for (size_t i = 0; i < nVoxels; ++i)
	{
		data[i] = slope * data[i] + intercept;
	}
	return trueActivities;
}

VolumeStat Volume::getStats()
{
	if (stats.dirty)
	{
		computeStats();
	}
	return stats;
}

void Volume::setOrigin(vec3 _origin) { 
	origin = _origin;
}

void Volume::setVoxelSize(float voxelWidth, float voxelHeight, float voxelDepth)
{
	voxelSize.x = voxelWidth;
	voxelSize.y = voxelHeight;
	voxelSize.z = voxelDepth;
	extent.x = width * voxelWidth;
	extent.y = height * voxelHeight;
	extent.z = depth * voxelDepth;
}

void Volume::scale(float scalingFactor)
{
	for (size_t i = 0; i < nVoxels; ++i)
	{
		hostPtr[i] *= scalingFactor;
	}
}

void Volume::scaleTo(const Volume& reference)
{
	float thisAverage = getAverageActivity();
	float otherAverage = reference.getAverageActivity();
	float scalingFactor = otherAverage / thisAverage;
	if (fabs(1.0f - scalingFactor) > 0.1f)
	{
		for (size_t i = 0; i < nVoxels; ++i)
		{
			hostPtr[i] *= scalingFactor;
		}
	}
}

float Volume::operator[](size_t index) const { return hostPtr[index]; }
float& Volume::operator[](size_t index) { return hostPtr[index]; }
float Volume::at(size_t x, size_t y, size_t z) const { return hostPtr[x + y * width + z * width * height]; }
float& Volume::at(size_t x, size_t y, size_t z) { return hostPtr[x + y * width + z * width * height]; }

void Volume::fill(float value) { std::fill(hostPtr, hostPtr + nVoxels, value); }

void Volume::flipAxially()
{
	float* tmpVol = new float[nVoxels];
	const size_t nVoxelsPerAxialSlice = width * height;
	for (size_t z = 0; z < depth; ++z)
	{
		memcpy(tmpVol + z * nVoxelsPerAxialSlice, hostPtr + (depth - z - 1) * nVoxelsPerAxialSlice, nVoxelsPerAxialSlice * sizeof(float));
	}
	delete[] hostPtr;
	hostPtr = tmpVol;

	if (isFlipped) origin.z += extent.z;
	else origin.z -= extent.z;
	isFlipped = !isFlipped;
}

void Volume::minmax(float& dmin, float& dmax)
{
	dmin = std::numeric_limits<float>::max();
	dmax = std::numeric_limits<float>::min(); // TODO: this returns FLT_MIN > 0

	for (size_t i = 0; i < width * height * depth; ++i)
	{
		dmin = std::min(dmin, hostPtr[i]);
		dmax = std::max(dmax, hostPtr[i]);
	}
}

#ifdef TEST
unsigned int poisson(double lambda)
{
	double elambda = exp(-lambda);
	int n = 0;

	if (lambda < 30) {
		double p = 1;
		do {
			n++;
			p *= (double)rand() / RAND_MAX;
		} while (p > elambda);
		n--;
	}
	else {
		// Rejection sampling method if lambda is too big
		double c = 0.767 - 3.36 / lambda;
		double beta = 3.1415 / sqrt(3.0 * lambda);
		double alpha = beta * lambda;
		double k = log(c) - lambda - log(beta);

		for (;;) {
			double u = (double)rand() / RAND_MAX;
			double x = (alpha - log((1.0 - u) / u)) / beta;
			n = floor(x + 0.5);
			if (n < 0) {
				continue;
			}
			double v = (double)rand() / RAND_MAX;
			double y = alpha - beta * x;
			double lhs = y + log(v / (1.0 + exp(y))) * 2;
			double lognfact = 0;
			for (int i = 1; i <= n; i++) { lognfact += log((double)i); }
			double rhs = k + n * log(lambda) - lognfact;
			if (lhs <= rhs) { break; }
		}
	}
	return n;
}

static float maxValue = 32768;

void Volume::reduceScanTime(std::shared_ptr<Volume> refVolume, int reductionPercent)
{
	printf("\nReduction: %d \n", reductionPercent);
	float reduction = reductionPercent / 100.0f * 0.05f;
	//	float reduction = reductionPercent / 100.0f;
	float oldAverage = 0, newAverage = 0;
	for (size_t i = 0; i < nVoxels; ++i)
	{
		float oldValue = fmax(refVolume->hostPtr[i], 0);
		oldAverage += oldValue / nVoxels;
	}
	const float uniPercent = 0.0f;

	for (size_t i = 0; i < nVoxels; ++i)
	{
		float oldValue = fmax(refVolume->hostPtr[i], 0);
		oldValue = oldValue * (1 - uniPercent) + uniPercent * oldAverage;
		float newValue = poisson(oldValue * reduction) / reduction;
//		if (newValue > maxValue) newValue = maxValue;

		hostPtr[i] = newValue;
		newAverage += newValue / nVoxels;
	}
	for (size_t i = 0; i < nVoxels; ++i)
	{
		hostPtr[i] *= oldAverage / newAverage;
	}
}

void Volume::computeSNR(Volume& inputVolume, const Volume& outputVolume,
	float& diff, float& inputSigma, float& outputSigma, float& inputSNR, float& outputSNR,
	float& inputPSNR, float& outputPSNR, float& inputSSIM, float& outputSSIM ) {
	float S2 = 0, inputS2 = 0, outputS2 = 0, inputN2 = 0.000001f, outputN2 = 0.000001f, inputAverage = 0, outputAverage = 0;
	float inputCov = 0, outputCov = 0, average = 0;
	diff = 0;

	int nRealVoxels = 0;
	for (size_t i = 0; i < nVoxels; ++i)
	{
		if (hostPtr[i] > 100) nRealVoxels++;
	}
	printf("nVoxels: %d, nRealVoxels: %d\n", nVoxels, nRealVoxels);

	for (size_t i = 0; i < nVoxels; ++i)
	{
		average += hostPtr[i] / nVoxels;
		S2 += hostPtr[i] * hostPtr[i] / nVoxels;
		inputS2 += inputVolume.hostPtr[i] * inputVolume.hostPtr[i] / nVoxels;
		outputS2 += outputVolume.hostPtr[i] * outputVolume.hostPtr[i] / nVoxels;
		inputN2 += (inputVolume.hostPtr[i] - hostPtr[i]) * (inputVolume.hostPtr[i] - hostPtr[i]) / nVoxels;
		outputN2 += (outputVolume.hostPtr[i] - hostPtr[i]) * (outputVolume.hostPtr[i] - hostPtr[i]) / nVoxels;
		inputCov += inputVolume.hostPtr[i] * hostPtr[i] / nVoxels;
		outputCov += outputVolume.hostPtr[i] * hostPtr[i] / nVoxels;
		diff += (outputVolume.hostPtr[i] - inputVolume.hostPtr[i]) * (outputVolume.hostPtr[i] - inputVolume.hostPtr[i]);
		inputAverage += inputVolume.hostPtr[i] / nVoxels;
		outputAverage += outputVolume.hostPtr[i] / nVoxels;
	}
	diff = sqrtf(diff);
	inputSigma = sqrtf(inputN2);
	outputSigma = sqrtf(outputN2);
	inputSNR = S2 / inputN2;
	outputSNR = S2 / outputN2;
	inputPSNR = 10 * log(maxValue * maxValue / inputN2) / log(10.0);
	outputPSNR = 10 * log(maxValue * maxValue / outputN2) / log(10.0);

	float scale = (float)nVoxels / nRealVoxels;
	inputCov = inputCov - inputAverage * average * scale;	// covariance with the reference
	outputCov = outputCov - outputAverage * average * scale;

	const float k1 = 0.01f, k2 = 0.03f;
	const float L = 256;
	float c1 = 0; // k1 * L * k1 * L;
	float c2 = 0; // k2 * L * k2 * L;

	inputSSIM = (2 * inputAverage * average * scale  + c1) * (2 * inputCov + c2) / 
		        (inputAverage * inputAverage * scale  + average * average * scale  + c1) /
			    (inputS2 - inputAverage * inputAverage * scale  + S2  - average * average  * scale + c2);
	outputSSIM = (2 * outputAverage * average * scale + c1) * (2 * outputCov + c2) /
			     (outputAverage * outputAverage * scale + average * average * scale  + c1) /
				 (outputS2 - outputAverage * outputAverage * scale + S2 - average * average * scale + c2);
	
	std::cout << "L2 Diff: " << diff << '\n';
	std::cout << "Input std: " << inputSigma << ", Output std: " << outputSigma << '\n';
	std::cout << "Input SNR: " << inputSNR << ", Output SNR: " << outputSNR << '\n';
	std::cout << "Input PSNR: " << inputPSNR << ", Output PSNR: " << outputPSNR << '\n';
	std::cout << "Input average: " << inputAverage << ", Output average: " << outputAverage << '\n';
	std::cout << "Input SSIM: " << inputSSIM << ", Output SSIM: " << outputSSIM << '\n';
}
#endif

void Volume::computeStats()
{
	std::vector<float> values(hostPtr, hostPtr + nVoxels);
	std::sort(values.begin(), values.end());	
	stats.vmin = values.front();
	stats.vmax = values.back();
	constexpr double percentile = 0.999;
	size_t iPercentile = static_cast<size_t>( static_cast<double>(nVoxels) * percentile );
	if (nVoxels > 0 && iPercentile >= nVoxels) iPercentile = nVoxels - 1; // handle tiny volumes
	stats.percentile = values[iPercentile];

	stats.histogram.resize(VolumeStat::histogramSize, 0.0f);
	for (const auto& v : values)
	{
		auto vclamped = std::max(v, 0.0f); // TODO: check data range
		int i = static_cast<int>(vclamped / stats.percentile * VolumeStat::histogramSize); // NOTE: for PET it is better to use this value instead of the real max
		i = std::min(i, VolumeStat::histogramSize - 1);
		++stats.histogram[i];
	}
	for (auto& hv : stats.histogram)
	{
		hv /= values.size();
	}
	
	stats.dirty = false;
}

void VolumeStat::reset()
{
	dirty = true;
	vmin = -std::numeric_limits<float>::infinity();
	vmax =  std::numeric_limits<float>::infinity();
	percentile = 0.0f;
	histogram.clear();
}
