#pragma once

#include "commonDefines.h"
#include <glm/glm.hpp>
#include <vector>
#include <limits>

#if APPLICATION == PET
#define nVolumes 1
#endif
#if APPLICATION == MR
#define nVolumes 3
#endif

inline glm::vec3 GLM(vec3 v) { return glm::vec3(v.x, v.y, v.z); }

struct VolumeStat
{
	bool dirty{ true };
	float vmin{ -std::numeric_limits<float>::infinity() };
	float vmax{  std::numeric_limits<float>::infinity() };
	float percentile{ 0.0f };
	std::vector<float> histogram;
	static constexpr int histogramSize = 256;
	
	void reset();
};

class Volume
{
	size_t width, height, depth;
	size_t nVoxels; // width * height * depth -> stored for conveniency
	vec3 origin; // in millimeter
	vec3 direction0, direction1;
	vec3 voxelSize; // in millimeter
	vec3 extent; // voxSize.x * width, voxSize.y * height, voxSize.z * depth -> stored for convenience
	float intercept; // offset of the intensity values, should be added to the data to get correct  values
	float slope; // slope of the intensity values, should be used to multiply the data to get correct values
	float *hostPtr;
	bool isFlipped = false;
	int axis = 2;
	std::vector<vec3> sliceOrigins; // origins of every slice
	VolumeStat stats;

	void computeStats();
public:
	Volume(size_t width, size_t height, size_t depth,
		vec3 origin = vec3(0, 0, 0), 
		vec3 direction0 = {1.0f, 0.0f, 0.0f}, vec3 direction1 = { 0.0f, 1.0f, 0.0f }, 
		vec3 voxelSize = vec3(1, 1, 1),
		float intercept = 0.0f, float slope = 1.0f,
		bool allocateDevice = true);
	Volume(const Volume& other);
	Volume& operator=(const Volume& other);
	~Volume();

	size_t getWidth() const;
	size_t getHeight() const;
	size_t getDepth() const;
	size_t getNumberOfVoxels() const;
	vec3 getOrigin() const;
	vec3 getOrigin(int z) const { return sliceOrigins[z]; }
	vec3 getDirection0() const { return direction0; }
	vec3 getDirection1() const { return direction1; }
	vec3 getVoxelSize() const;
	vec3 getExtent() const;
	float getIntercept() const;
	float getSlope() const;
	float* getHostPtr() const;
	float getSumOfActivity() const;
	float getAverageActivity() const;
	Volume* getTrueActivities() const;
	VolumeStat getStats();
	void setDirty() { stats.reset(); }
	void setOrigin(vec3 origin);
	void setVoxelSize(float voxelWidth, float voxelHeight, float voxelDepth);
	void scale(float scalingFactor);
	void scaleTo(const Volume& reference);
#ifdef TEST
	void reduceScanTime(std::shared_ptr<Volume> volume, int reductionPercent);
	void computeSNR(Volume& inputVolume, const Volume& outputVolume,
		float& diff, float& inputSigma, float& outputSigma, float& inputSNR, float& outputSNR,
		float& inputPSNR, float& outputPSNR, float& inputSSIM, float& outputSSIM);

#endif
	float operator[](size_t index) const;
	float& operator[](size_t index);
	float at(size_t x, size_t y, size_t z) const;
	float& at(size_t x, size_t y, size_t z);

	void fill(float value);

	void flipAxially();
	int  getAxis() const { return axis; }
	void setAxis(int ax) { axis = ax; }
	void setSliceOrigins(const std::vector<vec3>& _sliceOrigins) { sliceOrigins = _sliceOrigins; }
	vec3& getSliceOrigin(int depthIdx) { return sliceOrigins[depthIdx]; }

	void minmax(float& dmin, float& dmax);
	void print(std::string s) const;
};