#include <iostream>
#include <sstream>
#include <regex>
#include <memory>
#include <optional>

#include <nfd.h>

#include "GL/gl3w.h"
#include "GUIApp.h"
#include "gui/DebugOpenGL.hpp"

#include "imgui/imgui.h"
#include "imgui/imgui_impl_opengl3.h"
#include "imgui/imgui_impl_glfw.h"
#include "imgui_ext.h"

#include "media/raw.h"
#include "media/image.h"
#include "media/dicom.h"

#include "gui/shaders.h"
#include "gui/colormaps.h"

namespace PETFilter
{

void GUIApp::glfw_error_callback(int err, const char* description)
{
	std::cerr << "Glfw Error " << err << ": " << description << std::endl;
}

GUIApp::~GUIApp()
{

}

void GUIApp::Start(int argc, char* argv[])
{
	std::string workingDir;
	if (argc > 1)
	{
		workingDir = argv[1];
	}
	Init(workingDir);
	MessageLoop();
}

int GUIApp::Init(const std::string& workingDir)
{
	glfwSetErrorCallback(glfw_error_callback);
	if (!glfwInit()) return 1;

	const char* glsl_version = "#version 330";
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 2);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GLFW_TRUE);
	glfwWindowHint(GLFW_SAMPLES, 4);
	mWindow = glfwCreateWindow(800, 600, "Claritas PET_v1.01", NULL, NULL);
	if (mWindow == NULL) return -1;
	glfwMaximizeWindow(mWindow);
	glfwMakeContextCurrent(mWindow);

	bool err = gl3wInit() != 0;
	if (err)
	{
		fprintf(stderr, "Failed to initialize OpenGL loader!\n");
		return 1;
	}

	glEnable(GL_MULTISAMPLE);
	glfwSwapInterval(1); // Enable vsync
	glfwSetWindowUserPointer(mWindow, this);
	auto mouse_button_func = [](GLFWwindow* window, int button, int action, int mode)
	{
		static_cast<GUIApp*>(glfwGetWindowUserPointer(window))->glfw_mouse_button_callback(window, button, action, mode);
	};
	glfwSetMouseButtonCallback(mWindow, mouse_button_func);
	auto mouse_scroll_func = [](GLFWwindow* window, double xoffset, double yoffset)
	{
		static_cast<GUIApp*>(glfwGetWindowUserPointer(window))->glfw_mouse_scroll_callback(window, xoffset, yoffset);
	};
	glfwSetScrollCallback(mWindow, mouse_scroll_func);
	auto mouse_move_func = [](GLFWwindow* window, double xpos, double ypos)
	{
		static_cast<GUIApp*>(glfwGetWindowUserPointer(window))->glfw_mouse_position_callback(window, xpos, ypos);
	};
	glfwSetCursorPosCallback(mWindow, mouse_move_func);
	auto focus_func = [](GLFWwindow* window, int focusgiven)
	{
		static_cast<GUIApp*>(glfwGetWindowUserPointer(window))->glfw_windowfocus_callback(window, focusgiven);
	};
	glfwSetWindowFocusCallback(mWindow, focus_func);

#ifndef NDEBUG
	const GLubyte* GLVendor = glGetString(GL_VENDOR);
	const GLubyte* GLRenderer = glGetString(GL_RENDERER);
	std::cout << GLVendor << ": " << GLRenderer << std::endl;
#endif
	DebugOpenGL::init();
	DebugOpenGL::enableLowSeverityMessages(false);

	// Setup Dear ImGui context
	IMGUI_CHECKVERSION();
	ImGui::CreateContext();

	// Setup Platform/Renderer bindings
	ImGui_ImplGlfw_InitForOpenGL(mWindow, true);
	ImGui_ImplOpenGL3_Init(glsl_version);

	// Setup Style
	ImGui::StyleColorsClassic();

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glClearColor(mClearColor.x, mClearColor.y, mClearColor.z, mClearColor.w);
	glClear(GL_COLOR_BUFFER_BIT);
	glfwMakeContextCurrent(mWindow);
	glfwSwapBuffers(mWindow);

	mInputDataSet = std::make_shared<DataSet>();

	mInputTex.Init();

	m2DP.Create(vertexSource, fragmentSource, "outColor");
	m3DP.Create(vertexSourceVolumeRender, fragmentSourceVolumeRender, "outColor");
	mQuad.init();

	mDensityScales.emplace(VolumeRenderMode::ALPHA_BLEND, 2.0f);
	mDensityScales.emplace(VolumeRenderMode::ISOSURFACE, 1.0f);
	mDensityScales.emplace(VolumeRenderMode::MIP, 1.0f);
	mDensityScales.emplace(VolumeRenderMode::XRAY, 5.0f);

	mAlphaMapTex.Init();
	UpdateAlphaMap();
	glBindTexture(GL_TEXTURE_2D, mAlphaMapTex.textureId);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glBindTexture(GL_TEXTURE_2D, 0);

	mColorMaps.emplace("Grayscale", Texture());
	mColorMaps.emplace("Inverse Grayscale", Texture());
	mColorMaps.emplace("NIH", Texture());
	mColorMaps.emplace("Inverse NIH", Texture());
	for (auto& cmap : mColorMaps)
	{
		cmap.second.Init();
		mColorMapStr.append(cmap.first);
		mColorMapStr.append(1, '\0');
	}
	mColorMapStr.append(1, '\0');
	
	ColorMap::Init();
	mColorMaps["Grayscale"].Load3((int)ColorMap::Size, 1, ColorMap::grayScale.data());
	mColorMaps["Inverse Grayscale"].Load3((int)ColorMap::Size, 1, ColorMap::inverseGrayScale.data());
	mColorMaps["NIH"].Load3((int)ColorMap::Size, 1, ColorMap::NIH.data());
	mColorMaps["Inverse NIH"].Load3((int)ColorMap::Size, 1, ColorMap::inverseNIH.data());
	for (auto& cmap : mColorMaps)
	{
		glBindTexture(GL_TEXTURE_2D, cmap.second.textureId);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glBindTexture(GL_TEXTURE_2D, 0);
	}

	return 0;
}

void GUIApp::MessageLoop()
{
	while (!glfwWindowShouldClose(mWindow) && !mFinished)
	{
		glfwPollEvents();

		// Start the ImGui frame
		ImGui_ImplOpenGL3_NewFrame();
		ImGui_ImplGlfw_NewFrame();
		ImGui::NewFrame();

		//ImGuiIO& io = ImGui::GetIO();

		int display_w, display_h;
		glfwMakeContextCurrent(mWindow);
		glfwGetFramebufferSize(mWindow, &display_w, &display_h);

		MainMenu();
		SettingsPanel();
		ModalDialogs();

		// Rendering
		glViewport(0, 0, display_w, display_h);
		glClearColor(mClearColor.x, mClearColor.y, mClearColor.z, mClearColor.w);
		glClearDepth(1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		//ImGui::ShowDemoWindow();

		ImGui::EndFrame();
		ImGui::Render();

		switch (mRenderMode)
		{
		case RenderMode::SLICE:
			SliceRender(display_w, display_h);
			break;
		case RenderMode::VOLUMERENDER:
			VolumeRender(display_w, display_h);
			break;
		}

		ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

		glfwMakeContextCurrent(mWindow);
		glfwSwapBuffers(mWindow);
	}

	// Cleanup
	ImGui_ImplOpenGL3_Shutdown();
	ImGui_ImplGlfw_Shutdown();
	ImGui::DestroyContext();

	glfwDestroyWindow(mWindow);
	glfwTerminate();

}

void GUIApp::glfw_mouse_button_callback(GLFWwindow* window, int button, int action, int mods)
{
	ImGuiIO& io = ImGui::GetIO();
	if (io.WantCaptureMouse) return;

	if (mFirstClickAfterWindowFocus)
	{
		mFirstClickAfterWindowFocus = false;
		return;
	}

	if (button == GLFW_MOUSE_BUTTON_1 && action == GLFW_PRESS)
	{
		mDrag = true;
		glfwGetCursorPos(window, &mMouseX, &mMouseY);
	}
	else if (button == GLFW_MOUSE_BUTTON_1 && action == GLFW_RELEASE)
	{
		mDrag = false;
	}
}

void GUIApp::glfw_mouse_scroll_callback(GLFWwindow* window, double xoffset, double yoffset)
{
	ImGuiIO& io = ImGui::GetIO();
	if (io.WantCaptureMouse) return;
	float zoomFactor = 0.05f;

	switch (mRenderMode)
	{
	case RenderMode::SLICE:
		mInputImageScalerY *= 1.0f + copysignf(1.0f, (float)yoffset) * zoomFactor;
		mInputImageScalerX *= 1.0f + copysignf(1.0f, (float)yoffset) * zoomFactor;
		mFitImage = false;
		break;
	case RenderMode::VOLUMERENDER:
		float scaler = 1.0f - copysignf(1.0f, (float)yoffset) * zoomFactor;
		mCamera.Scroll(scaler);
		break;
	}
}

void GUIApp::glfw_mouse_position_callback(GLFWwindow* window, double xpos, double ypos)
{
	ImGuiIO& io = ImGui::GetIO();
	if (io.WantCaptureMouse) return;
	if (mDrag)
	{
		switch (mRenderMode)
		{
		case RenderMode::SLICE:
			mInputCenterX -= (int)(mMouseX - xpos);
			mInputCenterY -= (int)(mMouseY - ypos);
			break;
		case RenderMode::VOLUMERENDER:
			mCamera.Rotate((float)(mMouseX - xpos), (float)(mMouseY - ypos));
			break;
		}
		mMouseX = xpos;
		mMouseY = ypos;
	}
}

void GUIApp::glfw_windowfocus_callback(GLFWwindow* window, int focusgiven)
{
	if (focusgiven == GLFW_TRUE)
	{
		mFirstClickAfterWindowFocus = true;
	}
	else
	{
		mDrag = false;
	}
}

void GUIApp::MainMenu()
{
	bool fileNew{ false };
	bool fileOpen{ false };
	bool fileOpenDir{ false };
#if APPLICATION == MR
	bool fileOpenDirs[nVolumes] = { false };
#endif
	bool fileExit{ false };

	if (ImGui::BeginMainMenuBar())
	{
		if (ImGui::BeginMenu("File"))
		{
			ImGui::MenuItem("New", nullptr, &fileNew);
#if APPLICATION == PET
			ImGui::MenuItem("Open volume slice...", nullptr, &fileOpen);
			ImGui::MenuItem("Open volume volume dir...", nullptr, &fileOpenDir);
#endif
#if APPLICATION == MR
			ImGui::MenuItem("Open Axial volume dir...", nullptr, &fileOpenDirs[0]);
			ImGui::MenuItem("Open Coronal volume dir...", nullptr, &fileOpenDirs[1]);
			ImGui::MenuItem("Open Sagittal volume dir...", nullptr, &fileOpenDirs[2]);
#endif

			ImGui::Separator();
			ImGui::MenuItem("Exit", nullptr, &fileExit);
			ImGui::EndMenu();
		}
		ImGui::EndMainMenuBar();
	}

	if (fileNew)
	{
		mInputDataSet = std::make_shared<DataSet>();
	}

	if (fileOpen)
	{
		FileOpenDialog();
	}

	if (fileOpenDir)
	{
		FileOpenDirDialog();
	}

#if APPLICATION == MR
	for (int iVolume = 0; iVolume < nVolumes; iVolume++) {
		if (fileOpenDirs[iVolume])
		{
			mInputDataSet->SelectVolume(iVolume);
			FileOpenDirDialog();
		}
	}
#endif

	if (fileExit)
	{
		mFinished = true;
	}
}

void GUIApp::SettingsPanel()
{
	auto inputVolume = mInputDataSet->GetVolume();
	if (inputVolume == nullptr) return;

	int display_w, display_h;
	glfwGetFramebufferSize(mWindow, &display_w, &display_h);
	ImVec2 window_pos = ImVec2(ImGui::GetIO().DisplaySize.x, 25);
	ImVec2 window_pos_pivot = ImVec2(1.0f, 0.0f);
	ImGui::SetNextWindowPos(window_pos, ImGuiCond_Always, window_pos_pivot);

	ImGui::Begin("Settings", nullptr,
		ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoTitleBar | ImGuiWindowFlags_AlwaysAutoResize | ImGuiWindowFlags_NoResize);

	if (ImGui::Combo("Rendering mode", (int*)(&mRenderMode), "Slice (2D)\0Volume (3D)\0\0"))
	{
		mFitImage = true;
	}
	switch (mRenderMode)
	{
	case RenderMode::SLICE:
		SliceRenderOptions();
		break;
	case RenderMode::VOLUMERENDER:
		VolumeRenderOptions();
		break;
	//default:
	//	break;
	}
	
	if (ImGui::Button("Fit to window"))
	{
		mFitImage = true;
	}

	ImGui::Separator();
	ImGui::Text("Input voxel:  (%3.1f x %3.1f x %3.1f)",
		inputVolume->getVoxelSize().x, inputVolume->getVoxelSize().y, inputVolume->getVoxelSize().z);

	ImGui::End();
}

void GUIApp::SliceRenderOptions()
{
	if (ImGui::Combo("Interpolation", (int*)(&mDisplayFilter), "Nearest\0Linear\0\0"))
	{
		mInputTex.SetFiltering(GL_TEXTURE_MIN_FILTER, mDisplayFilter == DisplayFilter::LINEAR ? GL_LINEAR : GL_NEAREST);
		mInputTex.SetFiltering(GL_TEXTURE_MAG_FILTER, mDisplayFilter == DisplayFilter::LINEAR ? GL_LINEAR : GL_NEAREST);
	}

	ImGui::SliderFloat("Intensity scale", &mImageScale, 1.0f, 100.0f, "%.0f");
	mInputMax = (1.01f - (0.01f * mImageScale)) * mInputRange.y;

#if APPLICATION == PET
	if (ImGui::Combo("Plane", (int*)(&mDisplayAxis), "YZ\0XZ\0XY\0\0"))
	{
		mFitImage = true;
	}

	ImGui::SliderInt("PET Slice", &mInputImageSlice[(int)mDisplayAxis], 0, mInputImageSize[(int)mDisplayAxis] - 1);
#endif
}

void GUIApp::VolumeRenderOptions()
{
	int renderMode;
	switch (mVolumeRenderMode) {
	case VolumeRenderMode::MIP: renderMode = 0; break;
	case VolumeRenderMode::XRAY: renderMode = 1; break;
	case VolumeRenderMode::ALPHA_BLEND: renderMode = 2; break;
	case VolumeRenderMode::ISOSURFACE: renderMode = 3; break;
	}
	if (ImGui::Combo("3D Rendering method", (&renderMode), "MIP\0X-ray\0Alpha blend\0Iso-surface\0\0"))
	{
		//mFitImage = true;
		switch (renderMode) {
		case 0: mVolumeRenderMode = VolumeRenderMode::MIP; break;
		case 1: mVolumeRenderMode = VolumeRenderMode::XRAY; break;
		case 2: mVolumeRenderMode = VolumeRenderMode::ALPHA_BLEND; break;
		case 3: mVolumeRenderMode = VolumeRenderMode::ISOSURFACE; break;
		}
	}
#ifdef FULL_UI
	ImGui::SliderFloat("Render quality", &mRenderQuality, 0.3f, 1.0f);
#endif
	ImGui::SliderFloat("Density scale", &mDensityScales[mVolumeRenderMode], 0.01f, 10.0f);
	if (mVolumeRenderMode == VolumeRenderMode::ALPHA_BLEND)
	{
		ImGui::SliderFloat("Alpha scale", &mAlphaScale, 0.001f, 1.0f);
		ImGui::SliderFloat("Alpha border scale", &mBorderAlphaScale, 1.0f, 15.0f, "");
		//ImGui::SliderFloat("Alpha border width", &mBorderWidth, 0.001f, 0.1f); // the user is not supposed to modify this

		std::shared_ptr<Volume> volume;
		volume = mInputDataSet->GetVolume();
		//	printf("Display Mode: %d, textureId: %d\n", mDisplayMode, texture->textureId);

		std::vector<float> histogram;
		if (volume != nullptr)
		{
			auto stats = volume->getStats();
			histogram = stats.histogram;
		}

		constexpr float alphaDefault = 0.2f;
		if (ImGui::CurveEdit("Alpha curve", mAlphaCurve, histogram.empty() ? nullptr : &histogram, mAlphaScale / alphaDefault))
		{
			UpdateAlphaMap();
		}
		if (ImGui::Button("Reset Alpha curve"))
		{
			mAlphaCurve.clear();
			UpdateAlphaMap();
		}
		//ImGui::Image((ImTextureID)mAlphaMapTex.textureId, ImVec2(210, 20)); // only for debug
#ifdef FULL_UI
		ImGui::Checkbox("Apply lighting", &mApplyLighting);
#endif
	}
	ImGui::ColorPicker3("Background color", mBgColor, ImGuiColorEditFlags_InputRGB | ImGuiColorEditFlags_DisplayRGB | ImGuiColorEditFlags_NoSidePreview);
	
	if ((mVolumeRenderMode == VolumeRenderMode::ALPHA_BLEND && mApplyLighting) ||
		mVolumeRenderMode == VolumeRenderMode::ISOSURFACE)
	{
		if (ImGui::SliderFloat3("Light direction", mLightDir, -1, 1))
		{
			glm::vec3 lightDir; lightDir.x = mLightDir[0], lightDir.y = mLightDir[1], lightDir.z = mLightDir[2];
			if (glm::length(lightDir) < 0.001f)
			{
				lightDir = glm::vec3(1, 0, 0);
			}
			else
			{
				lightDir = glm::normalize(lightDir);
			}
			mLightDir[0] = lightDir.x, mLightDir[1] = lightDir.y, mLightDir[2] = lightDir.z;
		}
		ImGui::SliderFloat("Lighting strength", &mLightingStrength, 0, 1);
		ImGui::ColorPicker3("Light color", mLightColor, ImGuiColorEditFlags_InputRGB | ImGuiColorEditFlags_DisplayRGB | ImGuiColorEditFlags_NoSidePreview);
	}
	if (mVolumeRenderMode == VolumeRenderMode::ALPHA_BLEND || mVolumeRenderMode == VolumeRenderMode::MIP)
	{
		int cmapId = 0;
		for (auto& cmap : mColorMaps) 
		{
			if (cmap.first == mColorMap) break;
			++cmapId;
		}
		if (ImGui::Combo("Color map", &cmapId, mColorMapStr.c_str()))
		{
			int id = 0;
			for (auto& cmap : mColorMaps)
			{
				if (id == cmapId)
				{
					mColorMap = cmap.first;
					break;
				}
				++id;
			}
		}

		ImGui::Image((ImTextureID)(mColorMaps[mColorMap].textureId), ImVec2(210, 20));
	}
	if (/*mVolumeRenderMode == VolumeRenderMode::MIP ||*/
		mVolumeRenderMode == VolumeRenderMode::XRAY)
	{
		ImGui::Checkbox("Invert", &mInvert);
	}
	if (mVolumeRenderMode == VolumeRenderMode::ISOSURFACE)
	{
		ImGui::SliderFloat("Iso value", &mIsoValue, 0, 1);
	}
	ImGui::RangeSliderFloat("Range (x)", &mUvMin.x, &mUvMax.x, 0, 1, "%.2f - %.2f");
	ImGui::RangeSliderFloat("Range (y)", &mUvMin.y, &mUvMax.y, 0, 1, "%.2f - %.2f");
	ImGui::RangeSliderFloat("Range (z)", &mUvMin.z, &mUvMax.z, 0, 1, "%.2f - %.2f");
}

void GUIApp::ModalDialogs()
{
	if (mRawDimensionsDialog)
	{
		ImGui::OpenPopup("Raw data dimensions");
		bool opened = true;
		if (ImGui::BeginPopupModal("Raw data dimensions", &opened, ImGuiWindowFlags_AlwaysAutoResize))
		{
			ImGui::LabelText("##rddfs", "File size: %d bytes", fs::file_size(mInputDataPath));
			ImGui::LabelText("##rddds", "Load size: %d bytes", mRawDimensions[0] * mRawDimensions[1] * mRawDimensions[2] * mRawElementSize);

			ImGui::Separator();

			ImGui::InputInt("Width", &mRawDimensions[0], 1, 10);
			ImGui::InputInt("Height", &mRawDimensions[1], 1, 10);
			ImGui::InputInt("Depth", &mRawDimensions[2], 1, 10);

			for (int i = 0; i < 3; ++i)
			{
				if (mRawDimensions[i] < 1) mRawDimensions[i] = 1;
			}

			if (ImGui::Button("Load"))
			{
				if (mInputDataSet->LoadRaw(mInputDataPath, mRawDimensions[0], mRawDimensions[1], mRawDimensions[2]))
				{
					UpdateInputData();
					mRawDimensionsDialog = false;
					ImGui::CloseCurrentPopup();
				}
			}
			ImGui::EndPopup();
		}
	}

	if (mTxtDimensionsDialog)
	{
		ImGui::OpenPopup("Txt data dimensions");
		bool opened = true;
		if (ImGui::BeginPopupModal("Txt data dimensions", &opened, ImGuiWindowFlags_AlwaysAutoResize))
		{
			ImGui::LabelText("##txtdfs", "File: %d voxels", mNumberOfVoxels);
			ImGui::LabelText("##txtdds", "Load: %d voxels", mTxtDimensions[0] * mTxtDimensions[1] * mTxtDimensions[2]);

			ImGui::Separator();
			ImGui::InputInt("Width##txt", &mTxtDimensions[0], 1, 10);
			ImGui::InputInt("Height##txt", &mTxtDimensions[1], 1, 10);
			ImGui::InputInt("Depth##txt", &mTxtDimensions[2], 1, 10);

			for (int i = 0; i < 3; ++i)
			{
				if (mRawDimensions[i] < 1) mRawDimensions[i] = 1;
			}

			if (ImGui::Button("Load##txt"))
			{
				if (mInputDataSet->LoadTxt(mInputDataPath, mTxtDimensions[0], mTxtDimensions[1], mTxtDimensions[2]))
				{
					UpdateInputData();
					mTxtDimensionsDialog = false;
					ImGui::CloseCurrentPopup();
				}
			}
			ImGui::EndPopup();
		}
	}
}

void GUIApp::FileOpenDialog()
{
	const nfdchar_t* filterList{ "raw,dcm,dat,txt" };
	nfdchar_t* outPath{ nullptr };
	nfdresult_t result = NFD_OpenDialog(filterList, nullptr, &outPath);

	if (result == NFD_OKAY)
	{
		std::string mbPath(outPath);
		std::wstring wPath;
		int len = MultiByteToWideChar(CP_UTF8, 0, mbPath.c_str(), (int)mbPath.length(), NULL, 0);
		if (len > 0)
		{
			wPath.resize(len);
			MultiByteToWideChar(CP_UTF8, 0, mbPath.c_str(), (int)mbPath.length(), &wPath[0], len);
		}

		mInputDataPath = wPath;
		if (mInputDataPath.extension() == ".raw")
		{
			std::string fileName(mInputDataPath.filename().string());
			std::string pattern{ "(\\d+)[x,_](\\d+)[x,_](\\d+)"};
			std::regex rx(pattern);
			std::smatch match;
			if (std::regex_search(fileName, match, rx))
			{
				if (match.size() == 4)
				{
					mRawDimensions[0] = std::stoi(match[1]);
					mRawDimensions[1] = std::stoi(match[2]);
					mRawDimensions[2] = std::stoi(match[3]);
				}
			}
			mRawDimensionsDialog = true;
		}

		if (mInputDataPath.extension() == ".txt")
		{
			mTxtDimensions[0] = 1;
			mTxtDimensions[1] = 1;
			mTxtDimensions[2] = 1;
			mNumberOfVoxels = numberOfItems(mInputDataPath.generic_string());
			mTxtDimensionsDialog = true;
		}

		if (mInputDataPath.extension() == ".dcm" || mInputDataPath.extension() == "")
		{
			if (mInputDataSet->LoadDicom(mInputDataPath))
			{
				UpdateInputData();
			}
			else
			{
				std::cout << "Cannot load input file " << mInputDataPath.string() << std::endl;
			}
		}
		if (mInputDataPath.extension() == ".dat")
		{
			if (mInputDataSet->LoadDat(mInputDataPath))
			{
				UpdateInputData();
#ifdef TEST
				mReferenceVolume = std::make_unique<Volume>(*mInputDataSet->GetVolume());
				mInputDataSet->GetVolume()->reduceScanTime(mReferenceVolume, 2);
				UpdateInputData();
#endif
			}
			else
			{
				std::cout << "Cannot load input file " << mInputDataPath.string() << std::endl;
			}
		}
	}
	free(outPath);
}

void GUIApp::FileOpenDirDialog()
{
	nfdchar_t* outPath{ nullptr };
	nfdresult_t result = NFD_PickFolder(nullptr, &outPath);

	if (result == NFD_OKAY)
	{
		std::string mbPath(outPath);
		std::wstring wPath;
		int len = MultiByteToWideChar(CP_UTF8, 0, mbPath.c_str(), (int)mbPath.length(), NULL, 0);
		if (len > 0)
		{
			wPath.resize(len);
			MultiByteToWideChar(CP_UTF8, 0, mbPath.c_str(), (int)mbPath.length(), &wPath[0], len);
		}

		mInputDataPath = wPath;
		if (mInputDataSet->LoadDicomFiles(mInputDataPath))
		{
#ifdef TEST
			mReferenceVolume = std::make_unique<Volume>(*mInputDataSet->GetVolume());
			// Scan time reduction
			mInputDataSet->GetVolume()->reduceScanTime(mReferenceVolume, 20);
#endif
			UpdateInputData();
#ifdef PRINT
			mInputDataSet->GetVolume()->print("\nPET: ");
#endif
		}
		else
		{
			std::cout << "Cannot load input dir " << mInputDataPath.string() << std::endl;
		}
		
	}
	free(outPath);
}

void GUIApp::UpdateInputData()
{
	auto volume = mInputDataSet->GetVolume();
	volume->minmax(mInputRange.x, mInputRange.y);
	mInputMax = mInputRange.y;
	mImageScale = 1.0f;

	mInputImageSize = glm::ivec3(volume->getWidth(), volume->getHeight(), volume->getDepth());
	mInputImageSlice = glm::ivec3(volume->getWidth() / 2, volume->getHeight() / 2, volume->getDepth() / 2);
	mInputTex.Load((int)volume->getWidth(), (int)volume->getHeight(), (int)volume->getDepth(), glm::vec3(volume->getVoxelSize().x, volume->getVoxelSize().y, volume->getVoxelSize().z), volume->getHostPtr());
	mInputTex.SetFiltering(GL_TEXTURE_MIN_FILTER, mDisplayFilter == DisplayFilter::LINEAR ? GL_LINEAR : GL_NEAREST);
	mInputTex.SetFiltering(GL_TEXTURE_MAG_FILTER, mDisplayFilter == DisplayFilter::LINEAR ? GL_LINEAR : GL_NEAREST);

	mFitImage = true;
}

void GUIApp::SliceRender(int display_w, int display_h)
{
	if (mInputDataSet->GetVolume() != nullptr)
	{
#if APPLICATION == MR
		mInputTex.Load((int)mInputDataSet->GetVolume()->getWidth(),
			(int)mInputDataSet->GetVolume()->getHeight(),
			(int)mInputDataSet->GetVolume()->getDepth(),
			mInputDataSet->GetVolume()->getHostPtr());
#endif
		const glm::vec3& mInputOrigin = GLM(mInputDataSet->GetVolume()->getOrigin());
		const glm::vec3& mInputVoxelSize = GLM(mInputDataSet->GetVolume()->getVoxelSize());
		const glm::vec3& mInputExtent = GLM(mInputDataSet->GetVolume()->getExtent());

		int sourceAxis = mInputDataSet->GetVolume()->getAxis();

		if (mFitImage)
		{
			glm::ivec2 imageAxes{ 0, 1 };
			if ((int)mDisplayAxis != sourceAxis) {
				switch ((int)mDisplayAxis)
				{
				case 0:
					imageAxes = (sourceAxis == 1) ? glm::ivec2(2, 1) : glm::ivec2(1, 2);
					break;
				case 1:
					imageAxes = (sourceAxis == 2) ? glm::ivec2(0, 2) : glm::ivec2(2, 1);
					break;
				case 2:
					imageAxes = (sourceAxis == 0) ? glm::ivec2(2, 0) : glm::ivec2(0, 2);
					break;
				}
			}

			mInputImageScalerY = (float)mInputImageSize[imageAxes.y] / (float)display_h;
			mInputImageScalerX = mInputImageScalerY * mInputVoxelSize[imageAxes.y] / mInputVoxelSize[imageAxes.x];

			if (mInputImageSize[imageAxes.x] / mInputImageScalerX > display_w)
			{
				mInputImageScalerX = (float)mInputImageSize[imageAxes.x] / (float)display_w;
				mInputImageScalerY = mInputImageScalerX * mInputVoxelSize[imageAxes.x] / mInputVoxelSize[imageAxes.y];
			}

			mInputCenterX = (int)((display_w - (mInputImageSize[imageAxes.x] / mInputImageScalerX)) / 2.0f);
			mInputCenterY = (int)((display_h - (mInputImageSize[imageAxes.y] / mInputImageScalerY)) / 2.0f);

			mFitImage = false;
		}

		m2DP.Use();
		mInputTex.SetUniform(m2DP.getId(), "inputImage", 0);
		m2DP.SetUniform1f("inputScaler", 1.0f / mInputMax);
		m2DP.SetUniform1i("inputMode", mInputMode);
		m2DP.SetUniform2i("windowSize", display_w, display_h);
		m2DP.SetUniform2f("inputImageScaler", mInputImageScalerX, mInputImageScalerY);
		m2DP.SetUniform2i("inputCenter", mInputCenterX, mInputCenterY);
		m2DP.SetUniform1i("flipVertical", false);
		m2DP.SetUniform1i("flipHorizontal", false);
		float slice{ 0 };
		switch (mDisplayAxis)
		{
		case DisplayAxis::YZ:
			slice = mInputImageSlice.x / (float)mInputImageSize.x;
			break;
		case DisplayAxis::XZ:
			slice = mInputImageSlice.y / (float)mInputImageSize.y;
			break;
		case DisplayAxis::XY:
			slice = mInputImageSlice.z / (float)mInputImageSize.z;
			break;
		}
		m2DP.SetUniform1f("inputImageSlice", slice);
		m2DP.SetUniform1i("imageAxis", (int)mDisplayAxis);
		m2DP.SetUniform1i("sourceAxis", mInputDataSet->GetVolume()->getAxis());

		mQuad.render();
	}
}

void GUIApp::VolumeRender(int display_w, int display_h)
{
	Texture3D * texture = nullptr;
	std::shared_ptr<Volume> volume;

	if (mInputDataSet->GetVolume() == nullptr) return;
	volume = mInputDataSet->GetVolume();
	texture = &mInputTex;

	auto stats = volume->getStats();
	float vmax = stats.percentile;

	texture->SetFiltering(GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	texture->SetFiltering(GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	m3DP.Use();

	m3DP.SetUniform1i("renderMode", (int)mVolumeRenderMode);
	m3DP.SetUniform1f("quality", mRenderQuality);
	m3DP.SetUniform1f("aspect", (float)display_h / (float)display_w);
	vec3 volumeExtent(texture->width * texture->voxSize.x, texture->height * texture->voxSize.y, texture->depth * texture->voxSize.z);
	m3DP.SetUniform2f("volumeAspect", volumeExtent.y / volumeExtent.x, volumeExtent.z / volumeExtent.x);
	m3DP.SetUniform4f("volumeResolution", (float)texture->width, (float)texture->height, (float)texture->depth, 0);
	m3DP.SetUniform1f("densityScale", mDensityScales[mVolumeRenderMode]);
	m3DP.SetUniform1f("densityNormalize", 1.0f / vmax);
	m3DP.SetUniform1f("alphaScale", mAlphaScale);
	m3DP.SetUniform1f("borderAlphaScale", mBorderAlphaScale);
	m3DP.SetUniform1f("borderWidth", mBorderWidth);
	m3DP.SetUniformMat4("invViewMat", glm::inverse(mCamera.GetViewMat()));
	m3DP.SetUniform4f("bgColor", mBgColor[0], mBgColor[1], mBgColor[2], 0);
	m3DP.SetUniform4f("oLightDir", mLightDir[0], mLightDir[1], mLightDir[2], 0);
	m3DP.SetUniform4f("lightColor", mLightColor[0], mLightColor[1], mLightColor[2], 0);
	m3DP.SetUniform1i("shade", mApplyLighting ? 1 : 0);
	m3DP.SetUniform1f("lightingStrength", mLightingStrength);
	m3DP.SetUniform1f("isoValue", mIsoValue);
	m3DP.SetUniform1f("invert", mInvert ? 1.0f : 0.0f);
	m3DP.SetUniform3f("uvMin", mUvMin.x, mUvMin.y, mUvMin.z);
	m3DP.SetUniform3f("uvMax", mUvMax.x, mUvMax.y, mUvMax.z);

	texture->SetUniform(m3DP.getId(), "textureData", 0);
	mColorMaps[mColorMap].SetUniform(m3DP.getId(), "colorMap", 1);
	mAlphaMapTex.SetUniform(m3DP.getId(), "alphaMap", 2);

	mQuad.render();
}

void GUIApp::UpdateAlphaMap()
{
	constexpr size_t dsize = 4096;
	float data[dsize];
	if (mAlphaCurve.empty())
	{
		/*for (size_t i = 0; i < dsize; ++i)
		{
			data[i] = static_cast<float>(i) / (dsize - 1);
		}*/
		mAlphaCurve = { {0,0}, {1, 1} };
	}
	float xLast = 0;
	float dx = 1.0f / dsize;
	float yLast = mAlphaCurve.front()[1];
	//std::sort(mAlphaCurve.begin(), mAlphaCurve.end(), [](const auto& p1, const auto& p2) { return p1[0] < p2[0]; });
	mAlphaCurve.push_back({ 1.0f, mAlphaCurve.back()[1] });
	for (const auto& p : mAlphaCurve)
	{
		for (float x = xLast; x < p[0]; x += dx)
		{
			int ix = static_cast<int>(x * (dsize - 1));
			data[ix] = glm::mix(yLast, p[1], (static_cast<float>(ix) / (dsize - 1) - xLast) / (p[0] - xLast));
		}
		xLast = p[0];
		yLast = p[1];
	}
	mAlphaCurve.erase(--mAlphaCurve.end());
	mAlphaMapTex.Load(dsize, 1, data);
}

}
