#pragma once

#include <array>
#include <string>
#include <memory>
#include <filesystem>
#include <unordered_map>

#include <future>
#include <chrono>

#include <GL/gl3w.h>
#include <GLFW/glfw3.h>
#include <glm/glm.hpp>

#include "gui/GLGraphics.h"
#include "media/dataset.h"

#ifdef __cpp_lib_filesystem
	namespace fs = std::filesystem;
#elif __cpp_lib_experimental_filesystem
	namespace fs = std::experimental::filesystem;
#elif _MSC_VER <= 1900 && _MSC_VER > 1800
	namespace fs = std::experimental::filesystem;
#else
	#error "No filesystem support!"
#endif

namespace PETFilter
{

enum class DisplayFilter : int {
	NEAREST = 0,
	LINEAR
};

enum class DisplayAxis : int {
	YZ = 0,
	XZ,
	XY
};

enum class RenderMode : int {
	SLICE = 0,
	VOLUMERENDER
};

enum class VolumeRenderMode : int {
	ALPHA_BLEND = 0,
	XRAY,
	MIP,
	ISOSURFACE
};

class GUIApp
{
public:
	GUIApp() = default;
	GUIApp(const GUIApp&) = delete;
	GUIApp(const GUIApp&&) = delete;
	~GUIApp();

	void Start(int argc, char* argv[]);

private:
	int Init(const std::string& workingDir);
	void MessageLoop();

	void glfw_mouse_button_callback(GLFWwindow* window, int button, int action, int mods);
	void glfw_mouse_scroll_callback(GLFWwindow* window, double xoffset, double yoffset);
	void glfw_mouse_position_callback(GLFWwindow* window, double xpos, double ypos);
	void glfw_windowfocus_callback(GLFWwindow* window, int focusgiven);
	static void glfw_error_callback(int err, const char* description);

	void MainMenu();
	void SettingsPanel();
	void SliceRenderOptions();
	void VolumeRenderOptions();
	void ModalDialogs();
	void FileOpenDialog();
	void FileOpenDirDialog();
	void UpdateInputData();
	void SliceRender(int display_w, int display_h);
	void VolumeRender(int display_w, int display_h);
	void UpdateAlphaMap();

	bool mFinished{ false };
	GLFWwindow* mWindow{ nullptr };
	bool mFirstClickAfterWindowFocus{ true };
	bool mDrag{ false };
	double mMouseX{ 0 };
	double mMouseY{ 0 };
	int mInputCenterX{ 0 };
	int mInputCenterY{ 0 };
	bool mFitImage{ true };
	float mInputImageScalerX{ 1.0f };
	float mInputImageScalerY{ 1.0f };
	std::unordered_map<VolumeRenderMode, float> mDensityScales;
	RenderMode mRenderMode{ RenderMode::SLICE };
	float mBgColor[3]{ 0, 0, 0 };
	float mLightDir[3]{ 0, -1, 1 };
	float mLightColor[3]{ 1, 1, 1 };
	float mLightingStrength{ 0.5f };
	float mIsoValue{ 0.5f };
	float mAlphaScale{ 0.2f };
	float mBorderAlphaScale{ 2.0f };
	std::vector<std::array<float,2>> mAlphaCurve;
	Texture mAlphaMapTex;
	float mBorderWidth{ 0.005f };
	bool mApplyLighting{ false };
	bool mInvert{ false };
	float mRenderQuality{ 1.0f };
	bool mUseCullPlane{ false };
	glm::vec4 cullPlane{ glm::vec4(0,1,0,0) };
	glm::vec3 mUvMin{ glm::vec3(0,0,0) };
	glm::vec3 mUvMax{ glm::vec3(1,1,1) };
	VolumeRenderMode mVolumeRenderMode{ VolumeRenderMode::MIP };
	DisplayFilter mDisplayFilter{ DisplayFilter::LINEAR };
	DisplayAxis mDisplayAxis{ DisplayAxis::XY };
	int mInputMode{ 1 };
	std::unordered_map<std::string, Texture> mColorMaps;
	std::string mColorMap{ "Grayscale" };
	std::string mColorMapStr;

	glm::vec4 mClearColor{ 0.5f, 0.5f, 0.5f, 0.0f };
	glm::ivec3 mInputImageSize;
	glm::ivec3 mInputImageSlice;

	Texture3D mInputTex;
	glm::vec2 mInputRange{ 0, 0 };
	float mInputMax{ std::numeric_limits<float>::max() };
	float mImageScale{ 1.0f };
	float mSamplingDensity{ 1.0f };
	float mSamplingDensityMax{ 1.0f };

	GPUProgram m2DP;
	GPUProgram m3DP;
	Quad mQuad;
	OrbitingCamera mCamera;

	std::shared_ptr<DataSet> mInputDataSet;

	fs::path mInputDataPath;

	bool mRawDimensionsDialog{ false };
	glm::ivec3 mRawDimensions{ 1, 1, 1 };
	int mRawElementSize{ 4 };
	bool mTxtDimensionsDialog{ false };
	glm::ivec3 mTxtDimensions{ 1, 1, 1 };
	int mNumberOfVoxels = 0;
};

}