#include "imgui_ext.h"

#define IMGUI_DEFINE_MATH_OPERATORS
#include "imgui/imgui_internal.h"

#include <algorithm>
#include <iostream> // for debugging

using namespace ImGui;

bool RangeSliderBehavior(const ImRect& bb, ImGuiID id, float* v1, float* v2, float v_min, float v_max, const char* format, ImGuiSliderFlags flags, float& grab_sz, ImRect* out_grab_bb1, ImRect* out_grab_bb2)
{
    // Read imgui.cpp "API BREAKING CHANGES" section for 1.78 if you hit this assert.
    IM_ASSERT((flags == 1 || (flags & ImGuiSliderFlags_InvalidMask_) == 0) && "Invalid ImGuiSliderFlags flag!  Has the 'float power' argument been mistakenly cast to flags? Call function with ImGuiSliderFlags_Logarithmic flags instead.");

    ImGuiContext& g = *GImGui;
    if ((g.CurrentWindow->DC.ItemFlags & ImGuiItemFlags_ReadOnly) || (flags & ImGuiSliderFlags_ReadOnly))
        return false;

    const ImGuiStyle& style = g.Style;

    const ImGuiAxis axis = (flags & ImGuiSliderFlags_Vertical) ? ImGuiAxis_Y : ImGuiAxis_X;
    const bool is_decimal = true;
    const bool is_logarithmic = (flags & ImGuiSliderFlags_Logarithmic) && is_decimal;

    const float grab_padding = 2.0f;
    const float slider_sz = (bb.Max[axis] - bb.Min[axis]) - grab_padding * 2.0f;
    grab_sz = style.GrabMinSize;
    float v_range = (v_min < v_max ? v_max - v_min : v_min - v_max);
    if (!is_decimal && v_range >= 0)                                             // v_range < 0 may happen on integer overflows
        grab_sz = ImMax<float>((float)(slider_sz / (v_range + 1)), style.GrabMinSize);  // For integer sliders: if possible have the grab size represent 1 unit
    grab_sz = ImMin(grab_sz, slider_sz);
    const float slider_usable_sz = slider_sz - grab_sz;
    const float slider_usable_pos_min = bb.Min[axis] + grab_padding + grab_sz * 0.5f;
    const float slider_usable_pos_max = bb.Max[axis] - grab_padding - grab_sz * 0.5f;

    float logarithmic_zero_epsilon = 0.0f; // Only valid when is_logarithmic is true
    float zero_deadzone_halfsize = 0.0f; // Only valid when is_logarithmic is true
    if (is_logarithmic)
    {
        // When using logarithmic sliders, we need to clamp to avoid hitting zero, but our choice of clamp value greatly affects slider precision. We attempt to use the specified precision to estimate a good lower bound.
        const int decimal_precision = is_decimal ? ImParseFormatPrecision(format, 3) : 1;
        logarithmic_zero_epsilon = ImPow(0.1f, (float)decimal_precision);
        zero_deadzone_halfsize = (style.LogSliderDeadzone * 0.5f) / ImMax(slider_usable_sz, 1.0f);
    }

    // Process interacting with the slider
    bool value_changed = false;
    if (g.ActiveId == id)
    {
        bool set_new_value = false;
        float clicked_t = 0.0f;
        if (g.ActiveIdSource == ImGuiInputSource_Mouse)
        {
            if (!g.IO.MouseDown[0])
            {
                ClearActiveID();
            }
            else
            {
                const float mouse_abs_pos = g.IO.MousePos[axis];
                clicked_t = (slider_usable_sz > 0.0f) ? ImClamp((mouse_abs_pos - slider_usable_pos_min) / slider_usable_sz, 0.0f, 1.0f) : 0.0f;
                if (axis == ImGuiAxis_Y)
                    clicked_t = 1.0f - clicked_t;
                set_new_value = true;
            }
        }
        else if (g.ActiveIdSource == ImGuiInputSource_Nav)
        {
            if (g.ActiveIdIsJustActivated)
            {
                g.SliderCurrentAccum = 0.0f; // Reset any stored nav delta upon activation
                g.SliderCurrentAccumDirty = false;
            }

            const ImVec2 input_delta2 = GetNavInputAmount2d(ImGuiNavDirSourceFlags_Keyboard | ImGuiNavDirSourceFlags_PadDPad, ImGuiInputReadMode_RepeatFast, 0.0f, 0.0f);
            float input_delta = (axis == ImGuiAxis_X) ? input_delta2.x : -input_delta2.y;
            if (input_delta != 0.0f)
            {
                const int decimal_precision = is_decimal ? ImParseFormatPrecision(format, 3) : 0;
                if (decimal_precision > 0)
                {
                    input_delta /= 100.0f;    // Gamepad/keyboard tweak speeds in % of slider bounds
                    if (IsNavInputDown(ImGuiNavInput_TweakSlow))
                        input_delta /= 10.0f;
                }
                else
                {
                    if ((v_range >= -100.0f && v_range <= 100.0f) || IsNavInputDown(ImGuiNavInput_TweakSlow))
                        input_delta = ((input_delta < 0.0f) ? -1.0f : +1.0f) / (float)v_range; // Gamepad/keyboard tweak speeds in integer steps
                    else
                        input_delta /= 100.0f;
                }
                if (IsNavInputDown(ImGuiNavInput_TweakFast))
                    input_delta *= 10.0f;

                g.SliderCurrentAccum += input_delta;
                g.SliderCurrentAccumDirty = true;
            }

            float delta = g.SliderCurrentAccum;
            if (g.NavActivatePressedId == id && !g.ActiveIdIsJustActivated)
            {
                ClearActiveID();
            }
            else if (g.SliderCurrentAccumDirty)
            {
                clicked_t = ScaleRatioFromValueT<float, float, float>(ImGuiDataType_Float, *v1, v_min, v_max, is_logarithmic, logarithmic_zero_epsilon, zero_deadzone_halfsize);

                if ((clicked_t >= 1.0f && delta > 0.0f) || (clicked_t <= 0.0f && delta < 0.0f)) // This is to avoid applying the saturation when already past the limits
                {
                    set_new_value = false;
                    g.SliderCurrentAccum = 0.0f; // If pushing up against the limits, don't continue to accumulate
                }
                else
                {
                    set_new_value = true;
                    float old_clicked_t = clicked_t;
                    clicked_t = ImSaturate(clicked_t + delta);

                    // Calculate what our "new" clicked_t will be, and thus how far we actually moved the slider, and subtract this from the accumulator
                    float v_new = ScaleValueFromRatioT<float, float, float>(ImGuiDataType_Float, clicked_t, v_min, v_max, is_logarithmic, logarithmic_zero_epsilon, zero_deadzone_halfsize);
                    if (!(flags & ImGuiSliderFlags_NoRoundToFormat))
                        v_new = RoundScalarWithFormatT<float, float>(format, ImGuiDataType_Float, v_new);
                    float new_clicked_t = ScaleRatioFromValueT<float, float, float>(ImGuiDataType_Float, v_new, v_min, v_max, is_logarithmic, logarithmic_zero_epsilon, zero_deadzone_halfsize);

                    if (delta > 0)
                        g.SliderCurrentAccum -= ImMin(new_clicked_t - old_clicked_t, delta);
                    else
                        g.SliderCurrentAccum -= ImMax(new_clicked_t - old_clicked_t, delta);
                }

                g.SliderCurrentAccumDirty = false;
            }
        }

        if (set_new_value)
        {
            float v_new = ScaleValueFromRatioT<float, float, float>(ImGuiDataType_Float, clicked_t, v_min, v_max, is_logarithmic, logarithmic_zero_epsilon, zero_deadzone_halfsize);

            // Round to user desired precision based on format string
            if (!(flags & ImGuiSliderFlags_NoRoundToFormat))
                v_new = RoundScalarWithFormatT<float, float>(format, ImGuiDataType_Float, v_new);

            // Apply result
            if (*v1 != v_new || *v2 != v_new)
            {
                if (fabsf(*v1 - v_new) < fabsf(*v2 - v_new))
                {
                    *v1 = v_new;
                }
                else
                {
                    *v2 = v_new;
                }
                if (*v1 > * v2)
                {
                    float tmp = *v1;
                    *v1 = *v2;
                    *v2 = tmp;
                }
                value_changed = true;
            }
        }
    }

    if (slider_sz < 1.0f)
    {
        *out_grab_bb1 = ImRect(bb.Min, bb.Min);
        *out_grab_bb2 = ImRect(bb.Min, bb.Min);
    }
    else
    {
        // Output grab position so it can be displayed by the caller
        float grab_t1 = ScaleRatioFromValueT<float, float, float>(ImGuiDataType_Float, *v1, v_min, v_max, is_logarithmic, logarithmic_zero_epsilon, zero_deadzone_halfsize);
        if (axis == ImGuiAxis_Y)
            grab_t1 = 1.0f - grab_t1;
        const float grab_pos1 = ImLerp(slider_usable_pos_min, slider_usable_pos_max, grab_t1);
        if (axis == ImGuiAxis_X)
            *out_grab_bb1 = ImRect(grab_pos1 - grab_sz * 0.5f, bb.Min.y + grab_padding, grab_pos1 + grab_sz * 0.5f, bb.Max.y - grab_padding);
        else
            *out_grab_bb1 = ImRect(bb.Min.x + grab_padding, grab_pos1 - grab_sz * 0.5f, bb.Max.x - grab_padding, grab_pos1 + grab_sz * 0.5f);

        // Output grab position so it can be displayed by the caller
        float grab_t2 = ScaleRatioFromValueT<float, float, float>(ImGuiDataType_Float, *v2, v_min, v_max, is_logarithmic, logarithmic_zero_epsilon, zero_deadzone_halfsize);
        if (axis == ImGuiAxis_Y)
            grab_t2 = 1.0f - grab_t2;
        const float grab_pos2 = ImLerp(slider_usable_pos_min, slider_usable_pos_max, grab_t2);
        if (axis == ImGuiAxis_X)
            *out_grab_bb2 = ImRect(grab_pos2 - grab_sz * 0.5f, bb.Min.y + grab_padding, grab_pos2 + grab_sz * 0.5f, bb.Max.y - grab_padding);
        else
            *out_grab_bb2 = ImRect(bb.Min.x + grab_padding, grab_pos2 - grab_sz * 0.5f, bb.Max.x - grab_padding, grab_pos2 + grab_sz * 0.5f);
    }

    return value_changed;
}

bool ImGui::RangeSliderFloat(const char* label, float* v1, float* v2, float v_min, float v_max, const char* format, ImGuiSliderFlags flags)
{
    ImGuiWindow* window = GetCurrentWindow();
    if (window->SkipItems)
        return false;

    ImGuiContext& g = *GImGui;
    const ImGuiStyle& style = g.Style;
    const ImGuiID id = window->GetID(label);
    const float w = CalcItemWidth();

    const ImVec2 label_size = CalcTextSize(label, NULL, true);
    const ImRect frame_bb(window->DC.CursorPos, window->DC.CursorPos + ImVec2(w, label_size.y + style.FramePadding.y * 2.0f));
    const ImRect total_bb(frame_bb.Min, frame_bb.Max + ImVec2(label_size.x > 0.0f ? style.ItemInnerSpacing.x + label_size.x : 0.0f, 0.0f));

    ItemSize(total_bb, style.FramePadding.y);
    if (!ItemAdd(total_bb, id, &frame_bb))
        return false;

    // Default format string when passing NULL
    if (format == NULL)
        format = DataTypeGetInfo(ImGuiDataType_Float)->PrintFmt;

    // Tabbing or CTRL-clicking on Slider turns it into an input box
    const bool hovered = ItemHoverable(frame_bb, id);
    const bool temp_input_allowed = (flags & ImGuiSliderFlags_NoInput) == 0;
    bool temp_input_is_active = temp_input_allowed && TempInputIsActive(id);
    if (!temp_input_is_active)
    {
        const bool focus_requested = temp_input_allowed && FocusableItemRegister(window, id);
        const bool clicked = (hovered && g.IO.MouseClicked[0]);
        if (focus_requested || clicked || g.NavActivateId == id || g.NavInputId == id)
        {
            SetActiveID(id, window);
            SetFocusID(id, window);
            FocusWindow(window);
            g.ActiveIdUsingNavDirMask |= (1 << ImGuiDir_Left) | (1 << ImGuiDir_Right);
            if (temp_input_allowed && (focus_requested || (clicked && g.IO.KeyCtrl) || g.NavInputId == id))
            {
                temp_input_is_active = true;
                FocusableItemUnregister(window);
            }
        }
    }

    if (temp_input_is_active)
    {
        // Only clamp CTRL+Click input when ImGuiSliderFlags_AlwaysClamp is set
        const bool is_clamp_input = (flags & ImGuiSliderFlags_AlwaysClamp) != 0;
        return TempInputScalar(frame_bb, id, label, ImGuiDataType_Float, v1, format, is_clamp_input ? &v_min : NULL, is_clamp_input ? &v_max : NULL);
    }
    
    // Draw frame
    const ImU32 frame_col = GetColorU32(g.ActiveId == id ? ImGuiCol_FrameBgActive : g.HoveredId == id ? ImGuiCol_FrameBgHovered : ImGuiCol_FrameBg);
    RenderNavHighlight(frame_bb, id);
    RenderFrame(frame_bb.Min, frame_bb.Max, frame_col, true, g.Style.FrameRounding);

    // Slider behavior
    ImRect grab_bb1, grab_bb2;
    float grab_sz;
    //const bool value_changed = SliderBehavior(frame_bb, id, data_type, p_data, p_min, p_max, format, flags, &grab_bb);
    const bool value_changed = RangeSliderBehavior(frame_bb, id, v1, v2, v_min, v_max, format, flags, grab_sz, &grab_bb1, &grab_bb2);
    if (value_changed)
        MarkItemEdited(id);

    // Render grab
    if (grab_bb1.Max.x > grab_bb1.Min.x)
        window->DrawList->AddRectFilled(grab_bb1.Min, grab_bb1.Max, GetColorU32(g.ActiveId == id ? ImGuiCol_SliderGrabActive : ImGuiCol_SliderGrab), style.GrabRounding);

    if (grab_bb2.Max.x > grab_bb2.Min.x)
        window->DrawList->AddRectFilled(grab_bb2.Min, grab_bb2.Max, GetColorU32(g.ActiveId == id ? ImGuiCol_SliderGrabActive : ImGuiCol_SliderGrab), style.GrabRounding);

    ImRect connectLine(grab_bb1.Min, grab_bb2.Max);
    connectLine.Min.x += grab_sz;
    connectLine.Min.y += grab_sz * 0.35f;
    connectLine.Max.x -= grab_sz;
    connectLine.Max.y -= grab_sz * 0.35f;
    window->DrawList->AddRectFilled(connectLine.Min, connectLine.Max, GetColorU32(ImGuiCol_SliderGrab), style.GrabRounding);

    // Display value using user-provided display format so user can add prefix/suffix/decorations to the value.
    char value_buf[64];
    const char* value_buf_end = value_buf + ImFormatString(value_buf, IM_ARRAYSIZE(value_buf), format, *v1, *v2);
    RenderTextClipped(frame_bb.Min, frame_bb.Max, value_buf, value_buf_end, NULL, ImVec2(0.5f, 0.5f));

    if (label_size.x > 0.0f)
        RenderText(ImVec2(frame_bb.Max.x + style.ItemInnerSpacing.x, frame_bb.Min.y + style.FramePadding.y), label);

    IMGUI_TEST_ENGINE_ITEM_INFO(id, label, window->DC.ItemFlags);
    return value_changed;
}

ImVec2 TransformToRelativePos(const ImVec2& p, const ImRect& rect, bool invertY = true, bool clamp = true)
{
    ImVec2 result = (p - rect.Min);
    result.x /= rect.Max.x - rect.Min.x;
    result.y /= rect.Max.y - rect.Min.y;
    if (invertY) result.y = 1.0f - result.y;
    if (clamp)
    {
        result.x = std::clamp(result.x, 0.0f, 1.0f);
        result.y = std::clamp(result.y, 0.0f, 1.0f);
    }
    return result;
}

IMGUI_API bool ImGui::CurveEdit(const char* label, std::vector<std::array<float, 2>>& points, const std::vector<float>* bgHist, float histoScale)
{
    ImGuiWindow* window = GetCurrentWindow();
    if (window->SkipItems)
        return false;

    const ImGuiStyle& style = GetStyle();
    const ImGuiID id = window->GetID(label);
    ImGuiContext& g = *GImGui;

    constexpr float h = 100.0f;
    constexpr float border = 5.0f;
    const float w = CalcItemWidth() - 2*border;
    if (w < 0)
        return false;
    const ImVec2 label_size = CalcTextSize(label, NULL, true);
    const ImRect curve_bb(window->DC.CursorPos + ImVec2(border, border), window->DC.CursorPos + ImVec2(w + border, h + border));
    const ImRect frame_bb(window->DC.CursorPos, window->DC.CursorPos + ImVec2(w + 2 * border, h + 2 * border));
    const ImRect total_bb(frame_bb.Min, frame_bb.Max + ImVec2(label_size.x > 0.0f ? style.ItemInnerSpacing.x + label_size.x : 0.0f, 0.0f));

    ItemSize(total_bb, style.FramePadding.y);
    if (!ItemAdd(total_bb, id, &frame_bb))
        return false;

    const ImGuiIO& IO = GetIO();
    ImDrawList* DrawList = GetWindowDrawList();
    bool value_changed = false;

    RenderFrame(frame_bb.Min, frame_bb.Max, GetColorU32(ImGuiCol_FrameBg), true, style.FrameRounding);

    constexpr float pointSize = 5.0f;
    bool editorHovered = ItemHoverable(frame_bb, id);
    static bool dragging = false;
    static int iDragged = 0;
    if (editorHovered)
    {
        if (g.IO.MouseClicked[0])
        {
            //SetActiveID(id, window); // TODO: use it together with ClearActiveID
            FocusWindow(window);
        }

        auto hoveredIt = points.end();
        int iDraggedCurrent = 0;
        if (g.IO.MouseClicked[0] || g.IO.MouseClicked[1] || IsMouseDoubleClicked(0) || g.IO.MouseDown[0])
        {
            iDraggedCurrent = 0;
            for (auto it = points.begin(); it != points.end(); ++it, ++iDraggedCurrent)
            {
                const auto& p = *it;
                ImVec2 pCenter = curve_bb.Min + ImVec2(p[0] * w, (1.0f - p[1]) * h);
                bool hovered = ItemHoverable(ImRect(pCenter - ImVec2(pointSize, pointSize), pCenter + ImVec2(pointSize, pointSize)), id);
                if (hovered)
                {
                    hoveredIt = it;
                    break;
                }
            }
        }
        if (!g.IO.MouseDown[0])
            dragging = false;
        else if (hoveredIt != points.end())
        {
            dragging = true;
            iDragged = iDraggedCurrent;
        }

        // double left click: insert
        // TODO: do not allow points on the same X
        if (IsMouseDoubleClicked(0) 
            && hoveredIt == points.end()) // do not allow very close points
        {
            ImVec2 p = TransformToRelativePos(GetMousePos(), curve_bb);
            points.push_back({ p.x, p.y });
            std::sort(points.begin(), points.end(), [](const auto& p1, const auto& p2) { return p1[0] < p2[0]; });
            value_changed = true;
        }

        // left click: move
        ImVec2 mouseDelta = IO.MouseDelta;
        if (dragging && (mouseDelta.x != 0 || mouseDelta.y != 0) && iDragged < (int)points.size())
        {
            auto& p = points[iDragged];
            ImVec2 dp = TransformToRelativePos(mouseDelta, ImRect(0, 0, w, h), false, false);
            dp.y *= -1;
            if (iDragged > 0 && iDragged < (int)(points.size())-1)
                p[0] += dp.x;
            p[1] += dp.y;
            p[0] = std::clamp(p[0], 0.0f, 1.0f);
            p[1] = std::clamp(p[1], 0.0f, 1.0f);
            value_changed = true;
        }

        // right click: remove
        if (g.IO.MouseClicked[1] && hoveredIt != points.end() 
            && hoveredIt != points.begin() && hoveredIt != --points.end() ) // do not remove points at x = 0 and x = 1
        {
            points.erase(hoveredIt);
            value_changed = true;
        }
    }

    window->DrawList->AddRectFilled(curve_bb.Min, curve_bb.Max, GetColorU32(ImGuiCol_PlotHistogram), style.GrabRounding);
    for (size_t i = 0; i < points.size(); ++i)
    {
        const auto& p = points[i];
        ImVec2 pCenter = curve_bb.Min + ImVec2(p[0]*w, (1.0f - p[1])*h);
        bool hovered = ItemHoverable(ImRect(pCenter - ImVec2(pointSize, pointSize), pCenter + ImVec2(pointSize, pointSize)), id);
        hovered = hovered || (dragging && i == iDragged);
        window->DrawList->AddRectFilled(pCenter - ImVec2(pointSize, pointSize), pCenter + ImVec2(pointSize, pointSize), GetColorU32(hovered ? ImGuiCol_PlotHistogramHovered : ImGuiCol_PlotLines), style.GrabRounding);
        if (i > 0)
        {
            const auto& p2 = points[i-1];
            ImVec2 p2Center = curve_bb.Min + ImVec2(p2[0] * w, (1.0f - p2[1]) * h);
            window->DrawList->AddLine(pCenter, p2Center, GetColorU32(ImGuiCol_PlotLines));
        }
    }

    if (bgHist)
    {
        float histoMax = *std::max_element(bgHist->begin(), bgHist->end());
        float rectWidth = 1.0f / bgHist->size() * histoScale;
        for (size_t i = 0; i < bgHist->size(); ++i)
        {
            ImVec2 rectMin(curve_bb.Min.x + static_cast<float>(i)* rectWidth* w, curve_bb.Max.y - (*bgHist)[i] * h / histoMax);
            ImVec2 rectMax(rectMin.x + rectWidth * w, curve_bb.Max.y);
            window->DrawList->AddRectFilled(rectMin, rectMax, GetColorU32(ImGuiCol_FrameBgActive), style.GrabRounding);
        }
    }

    if (label_size.x > 0.0f)
        RenderText(ImVec2(frame_bb.Max.x + style.ItemInnerSpacing.x, frame_bb.Min.y + style.FramePadding.y), label);

    return value_changed;
}
