#include "dataset.h"

#include "raw.h"
#include "dicom.h"


namespace PETFilter
{
	bool DataSet::LoadDat(const fs::path& path)
	{
		//		int width = 192, height = 96, depth = 498;
		int width = 256, height = 256, depth = 128;
		vec3 origin = { 0.0f, 0.0f, 0.0f };
		vec3 spacing = { 1.0f, 1.0f, 1.0f };
		vec3 directions[2] = { {1.0f, 0.0f, 0.0f}, {0.0f, 1.0f, 0.0f} };
		float slope = 1, intercept = 0;

		SetVolume(std::make_shared<Volume>(width, height, depth, origin, directions[0], directions[1], spacing, slope, intercept));

		FILE * file = fopen(path.string().c_str(), "r");
		if (file == nullptr) return false;
		std::vector<unsigned char> imageData(width * height * depth);
		fread(&imageData[0], width * height * depth, 1, file);

		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				for (int z = 0; z < depth; z++) {
					int i1 = x + width * y + width * height * z;
					int i2 = x + width * y + width * height * (depth - 1 - z);
					GetVolume()->getHostPtr()[i1] = (float)imageData[i2];
				}
			}
		}

		mRepresentativeFilePath = path;
		return true;
	}

	bool DataSet::LoadDicom(const fs::path& path)
	{
		int width, height, channels;
		vec3 origin  = { 0.0f, 0.0f, 0.0f };
		vec3 spacing = { 1.0f, 1.0f, 1.0f };
		vec3 directions[2] = { {1.0f, 0.0f, 0.0f}, {0.0f, 1.0f, 0.0f} };
		float slope, intercept;
		std::vector<float> imageData;

		if (loadDicomImage(path.string(), width, height, channels, origin, directions, spacing, slope, intercept, imageData))
		{
			if (channels == 3) return false;

			SetVolume(std::make_shared<Volume>(width, height, 1, origin, directions[0], directions[1], spacing, slope, intercept));
			std::memcpy(reinterpret_cast<char*>(GetVolume()->getHostPtr()), imageData.data(), width * height * sizeof(float));
			mRepresentativeFilePath = path;
			return true;
		}
		return false;
	}

	bool DataSet::LoadDicomFiles(const fs::path& path)
	{
		std::vector<fs::path> files;

		for (const auto& f : fs::directory_iterator(path))
		{
//			if (f.path().extension() == ".dcm" || f.path().extension() == "")
			if (f.path().extension() == ".dcm")
			{
				files.push_back(f.path());
//				std::cout << f.path() << '\n';
			}
		}

		if (!files.empty()) {
			int baseWidth, baseHeight, baseChannels;
			int width, height, channels;
			vec3 origin = { 0.0f, 0.0f, 0.0f };
			vec3 minOrigin = { 0.0f, 0.0f, 0.0f };
			vec3 maxOrigin = { 0.0f, 0.0f, 0.0f };
			vec3 spacing   = { 1.0f, 1.0f, 1.0f };
			vec3 directions[3] = { {1.0f, 0.0f, 0.0f}, {0.0f, 1.0f, 0.0f}, {0.0f, 0.0f, 1.0f}, };
			float slope, intercept;
			std::vector<std::vector<float>> imageData;
			imageData.resize(files.size());
			std::vector<vec3> sliceOrigins;
			sliceOrigins.resize(files.size());

			if (!loadDicomImage(files[0].string(), baseWidth, baseHeight, baseChannels, origin, directions, spacing, slope, intercept, imageData[0]) || baseChannels == 3)
			{
				return false;
			}
			mRepresentativeFilePath = files[0];
			sliceOrigins[0] = origin;
			directions[2] = cross(directions[0], directions[1]);

			int nRealLayers = 1;
			float maxDepth = 0, minDepth = 0;
			int minDepthLayerIdx = 0;

			for (int i = 1; i < files.size(); ++i)
			{
				if (loadDicomImage(files[i].string(), width, height, channels, origin, directions, spacing, slope, intercept, imageData[nRealLayers]))
				{
					if (width != baseWidth || height != baseHeight || channels != baseChannels) {
						continue;
					}
					sliceOrigins[nRealLayers] = origin;
					float depth = dot(origin - sliceOrigins[0], directions[2]);
					if (depth > maxDepth) maxDepth = depth;
					if (depth < minDepth) {
						minDepth = depth;
						minDepthLayerIdx = nRealLayers;
					}
					nRealLayers++;
				}
			}

			origin = sliceOrigins[minDepthLayerIdx];

			if (nRealLayers > 0)
			{
				float voxelDepth = (maxDepth - minDepth) / (nRealLayers - 1.0f);

				SetVolume(std::make_shared<Volume>(baseWidth, baseHeight, nRealLayers, origin, directions[0], directions[1]));
				GetVolume()->setVoxelSize(spacing.x, spacing.y, (voxelDepth > 0) ? voxelDepth : spacing.z);
				if (fabs(directions[2].z) > fabs(directions[2].x) && fabs(directions[2].z) > fabs(directions[2].y))
					GetVolume()->setAxis(2);
				else if (fabs(directions[2].y) > fabs(directions[2].x))
					GetVolume()->setAxis(1);
				else
					GetVolume()->setAxis(0);

#ifdef PRINT
				printf("Axis: %d, VoxelSize: %f, %f, %f\n", GetVolume()->getAxis(), GetVolume()->getVoxelSize().x, GetVolume()->getVoxelSize().y, GetVolume()->getVoxelSize().z);
				printf("Extent: %f, %f, %f\n", GetVolume()->getExtent().x, GetVolume()->getExtent().y, GetVolume()->getExtent().z);

				std::cout << width << " x " << height
					<< ", org: " << origin.x << ", " << origin.y << ", " << origin.z
					<< ", sp: " << spacing.x << ", " << spacing.y << ", " << spacing.z << '\n';
				std::cout << "direction0 " << directions[0].x << ", " << directions[0].y << ", " << directions[0].z << '\n';
				std::cout << "direction1 " << directions[1].x << ", " << directions[1].y << ", " << directions[1].z << '\n';
#endif

				auto sliceSize = baseWidth * baseHeight * sizeof(float);
				for (int i = 0; i < nRealLayers; ++i) {
					if (voxelDepth > 0) {
						double depthCoordValue = dot(sliceOrigins[i] - origin, directions[2]);
						if (depthCoordValue < 0) {
//							printf("Negative depth: %f\n", depthCoordValue);
							depthCoordValue = 0;
						}
						const float eps = 0.01f;
						double slice = depthCoordValue / voxelDepth + eps;
						int sliceIdx = (int)slice;
//						printf("Slice: %d, %d, %f, %f\n", i, sliceIdx, depthCoordValue, slice);
						if (sliceIdx > nRealLayers - 1) sliceIdx = nRealLayers - 1;
						// This also sorts slices with increasing depthCoordValue
						std::memcpy(reinterpret_cast<char*>(GetVolume()->getHostPtr()) + (sliceIdx * sliceSize), imageData[i].data(), sliceSize);
						// slice origins are also sorted
						GetVolume()->getSliceOrigin(sliceIdx) = sliceOrigins[i];
					}
					else {
						std::memcpy(reinterpret_cast<char*>(GetVolume()->getHostPtr()) + (i * sliceSize), imageData[i].data(), sliceSize);
					}
				}
//				for (int i = 0; i < nRealLayers; i++)
//					printf("Slice Origin: %f, %f, %f\n", 
//						    GetVolume()->getSliceOrigin(i).x, GetVolume()->getSliceOrigin(i).y, GetVolume()->getSliceOrigin(i).z);

/*				if (voxelDepth < 0)
				{
					GetVolume()->flipAxially();
				}
*/
			}
			else
			{
				GetVolume()->setVoxelSize(spacing.x, spacing.y, spacing.z);
			}
			return true;
		}
		return false;
	}

	bool DataSet::LoadDynamicDicomFiles(const fs::path& path, int frameIdx) // frameIdx == -1 -> all
	{
		std::vector<fs::path> files;
		int inFrameIdx = 0;
		int curInFrameIdx = 0, curFrameIdx = 0;

		for (const auto& f : fs::directory_iterator(path))
		{
			if (f.path().extension() == ".dcm" || f.path().extension() == "")
			{
				std::cout << f.path() << '\n';
				const char * fileName = f.path().string().c_str();
				int i = strlen(fileName) - 1;
				for (; i >= 0; i--) if (fileName[i] == '_') break;
				int curInFrameIdx = atoi(&fileName[i + 1]);

				if (curFrameIdx == frameIdx) {
					if (curInFrameIdx < inFrameIdx) break;
					files.push_back(f.path());
				}
				else {
					if (curInFrameIdx < inFrameIdx) {
						curInFrameIdx = 0;
						curFrameIdx++;
						if (curFrameIdx == frameIdx) files.push_back(f.path());
					}
				}
				inFrameIdx = curInFrameIdx;
			}
		}

		if (!files.empty()) {
			int baseWidth, baseHeight, baseChannels;
			int width, height, channels;
			vec3 origin = { 0.0f, 0.0f, 0.0f };
			vec3 minOrigin = { 0.0f, 0.0f, 0.0f };
			vec3 maxOrigin = { 0.0f, 0.0f, 0.0f };
			vec3 spacing = { 1.0f, 1.0f, 1.0f };
			vec3 directions[2] = { {1.0f, 0.0f, 0.0f}, {0.0f, 1.0f, 0.0f} };
			float slope, intercept;
			std::vector<std::vector<float>> imageData;
			imageData.resize(files.size());
			std::vector<vec3> sliceOrigins;
			sliceOrigins.resize(files.size());

			if (!loadDicomImage(files[0].string(), baseWidth, baseHeight, baseChannels, origin, directions, spacing, slope, intercept, imageData[0]) || baseChannels == 3)
			{
				return false;
			}
			mRepresentativeFilePath = files[0];
			sliceOrigins[0] = minOrigin = maxOrigin = origin;

			//			float firstOriginZ = origin[2];
			int filesInFrame = 1;

			for (int i = 1; i < files.size(); ++i)
			{
				if (loadDicomImage(files[i].string(), width, height, channels, origin, directions, spacing, slope, intercept, imageData[i]))
				{
					if (width != baseWidth || height != baseHeight || channels != baseChannels)
					{
						continue;
					}
					sliceOrigins[i] = origin;

					filesInFrame++;
					if (minOrigin.x > origin.x) minOrigin.x = origin.x;
					if (maxOrigin.x < origin.x) maxOrigin.x = origin.x;
					if (minOrigin.y > origin.y) minOrigin.y = origin.y;
					if (maxOrigin.y < origin.y) maxOrigin.y = origin.y;
					if (minOrigin.z > origin.z) minOrigin.z = origin.z;
					if (maxOrigin.z < origin.z) maxOrigin.z = origin.z;
				}
			}

			if (filesInFrame > 0)
			{
				//				float lastOriginZ = origin[2];
				//				float voxelDepth = (lastOriginZ - firstOriginZ) / (files.size() - 1.0f);
				int depthCoord = 0;
				float voxelDepth = (maxOrigin.x - minOrigin.x);
				if (voxelDepth < maxOrigin.y - minOrigin.y) {
					voxelDepth = maxOrigin.y - minOrigin.y;
					depthCoord = 1;
				}
				if (voxelDepth < maxOrigin.z - minOrigin.z) {
					voxelDepth = maxOrigin.z - minOrigin.z;
					depthCoord = 2;
				}
				voxelDepth /= (filesInFrame - 1.0f);

				SetVolume(std::make_shared<Volume>(baseWidth, baseHeight, filesInFrame, minOrigin, directions[0], directions[1]));
				GetVolume()->setVoxelSize(spacing.x, spacing.y, (voxelDepth > 0) ? voxelDepth : spacing.z);
				GetVolume()->setAxis(depthCoord);

				auto sliceSize = width * height * sizeof(float);
				for (int i = 0; i < filesInFrame; ++i) {
					if (voxelDepth > 0) {
						float depthCoordValue = (depthCoord == 0) ? (sliceOrigins[i].x - minOrigin.x) :
							((depthCoord == 1) ? (sliceOrigins[i].y - minOrigin.y) :
							(sliceOrigins[i].z - minOrigin.z));

						float slice = depthCoordValue / voxelDepth;
						int sliceIdx = (int)slice;
						if (sliceIdx > filesInFrame - 1) sliceIdx = filesInFrame - 1;
						std::memcpy(reinterpret_cast<char*>(GetVolume()->getHostPtr()) + (sliceIdx * sliceSize), imageData[i].data(), sliceSize);
					}
					else {
						std::memcpy(reinterpret_cast<char*>(GetVolume()->getHostPtr()) + (i * sliceSize), imageData[i].data(), sliceSize);
					}
				}

				if (voxelDepth < 0)
				{
					GetVolume()->flipAxially();
				}
			}
			else
			{
				GetVolume()->setVoxelSize(spacing.x, spacing.y, spacing.z);
			}
			return true;
		}
		return false;
	}

	bool DataSet::LoadDynamicDicomFiles(const fs::path& path) // frameIdx == -1 -> all
	{
		std::vector<fs::path> files;
		int inFrameIdx = 0, frameIdx = 0;
		int curInFrameIdx = 0, curFrameIdx = 0;

		for (const auto& f : fs::directory_iterator(path))
		{
			if (f.path().extension() == ".dcm" || f.path().extension() == "")
			{
				std::cout << f.path() << '\n';
				const char * fileName = f.path().string().c_str();
				int i = strlen(fileName) - 1;
				for (; i >= 0; i--) if (fileName[i] == '_') break;
				int curInFrameIdx = atoi(&fileName[i + 1]);

				if (curFrameIdx == frameIdx) {
					if (curInFrameIdx < inFrameIdx) break;
					files.push_back(f.path());
				}
				else {
					if (curInFrameIdx < inFrameIdx) {
						curInFrameIdx = 0;
						curFrameIdx++;
						if (curFrameIdx == frameIdx) files.push_back(f.path());
					}
				}
				inFrameIdx = curInFrameIdx;
			}
		}

		if (!files.empty()) {
			int baseWidth, baseHeight, baseChannels;
			int width, height, channels;
			vec3 origin = { 0.0f, 0.0f, 0.0f };
			vec3 minOrigin = { 0.0f, 0.0f, 0.0f };
			vec3 maxOrigin = { 0.0f, 0.0f, 0.0f };
			vec3 spacing = { 1.0f, 1.0f, 1.0f };
			vec3 directions[2] = { {1.0f, 0.0f, 0.0f}, {0.0f, 1.0f, 0.0f} };
			float slope, intercept;
			std::vector<std::vector<float>> imageData;
			imageData.resize(files.size());
			std::vector<vec3> sliceOrigins;
			sliceOrigins.resize(files.size());

			if (!loadDicomImage(files[0].string(), baseWidth, baseHeight, baseChannels, origin, directions, spacing, slope, intercept, imageData[0]) || baseChannels == 3)
			{
				return false;
			}
			mRepresentativeFilePath = files[0];
			sliceOrigins[0] = minOrigin = maxOrigin = origin;

			//			float firstOriginZ = origin[2];
			int filesInFrame = 1;

			for (int i = 1; i < files.size(); ++i)
			{
				if (loadDicomImage(files[i].string(), width, height, channels, origin, directions, spacing, slope, intercept, imageData[i]))
				{
					if (width != baseWidth || height != baseHeight || channels != baseChannels)
					{
						continue;
					}
					sliceOrigins[i] = origin;
					//					for (int j = 0; j < i; j++) if (sliceOrigins[j] == sliceOrigins[i]) break;
					filesInFrame++;
					if (minOrigin.x > origin.x) minOrigin.x = origin.x;
					if (maxOrigin.x < origin.x) maxOrigin.x = origin.x;
					if (minOrigin.y > origin.y) minOrigin.y = origin.y;
					if (maxOrigin.y < origin.y) maxOrigin.y = origin.y;
					if (minOrigin.z > origin.z) minOrigin.z = origin.z;
					if (maxOrigin.z < origin.z) maxOrigin.z = origin.z;
				}
			}

			if (filesInFrame > 0)
			{
				//				float lastOriginZ = origin[2];
				//				float voxelDepth = (lastOriginZ - firstOriginZ) / (files.size() - 1.0f);
				int depthCoord = 0;
				float voxelDepth = (maxOrigin.x - minOrigin.x);
				if (voxelDepth < maxOrigin.y - minOrigin.y) {
					voxelDepth = maxOrigin.y - minOrigin.y;
					depthCoord = 1;
				}
				if (voxelDepth < maxOrigin.z - minOrigin.z) {
					voxelDepth = maxOrigin.z - minOrigin.z;
					depthCoord = 2;
				}
				voxelDepth /= (filesInFrame - 1.0f);

				SetVolume(std::make_shared<Volume>(baseWidth, baseHeight, filesInFrame, minOrigin, directions[0], directions[1]));
				GetVolume()->setVoxelSize(spacing.x, spacing.y, (voxelDepth > 0) ? voxelDepth : spacing.z);

				auto sliceSize = width * height * sizeof(float);
				for (int i = 0; i < filesInFrame; ++i) {
					if (voxelDepth > 0) {
						float depthCoordValue = (depthCoord == 0) ? (sliceOrigins[i].x - minOrigin.x) :
							((depthCoord == 1) ? (sliceOrigins[i].y - minOrigin.y) :
							(sliceOrigins[i].z - minOrigin.z));

						float slice = depthCoordValue / voxelDepth;
						int sliceIdx = (int)slice;
						if (sliceIdx > filesInFrame - 1) sliceIdx = filesInFrame - 1;
						std::memcpy(reinterpret_cast<char*>(GetVolume()->getHostPtr()) + (sliceIdx * sliceSize), imageData[i].data(), sliceSize);
					}
					else {
						std::memcpy(reinterpret_cast<char*>(GetVolume()->getHostPtr()) + (i * sliceSize), imageData[i].data(), sliceSize);
					}
				}

				if (voxelDepth < 0)
				{
					GetVolume()->flipAxially();
				}
			}
			else
			{
				GetVolume()->setVoxelSize(spacing.x, spacing.y, spacing.z);
			}
			return true;
		}
		return false;
	}


	bool DataSet::LoadRaw(const fs::path& path, const int width, const int height, const int depth)
	{
		SetVolume(std::make_shared<Volume>(width, height, depth));
		if (GetVolume()->getHostPtr() == nullptr) return false;
		loadFromVox(path.generic_string(), *GetVolume());
		mRepresentativeFilePath = path;
		return true;
	}

	bool DataSet::LoadTxt(const fs::path& path, const int width, const int height, const int depth)
	{
		SetVolume(std::make_shared<Volume>(width, height, depth));
		if (GetVolume()->getHostPtr() == nullptr) return false;
		loadFromTxt(path.generic_string(), *GetVolume());
		mRepresentativeFilePath = path;
		return true;
	}
}