
#include <numeric>
#include <vector>

#include <gdcmImageReader.h>
#include <gdcmImageWriter.h>
#include <gdcmDirectory.h>
#include <gdcmDataSet.h>

#include "dicom.h"

#include <filesystem>
#ifdef __cpp_lib_filesystem
	namespace fs = std::filesystem;
#elif __cpp_lib_experimental_filesystem
	namespace fs = std::experimental::filesystem;
#elif _MSC_VER <= 1900 && _MSC_VER > 1800
	namespace fs = std::experimental::filesystem;
#else
	#error "No filesystem support!"
#endif

namespace PETFilter
{

bool loadDicomImage(const std::string& fileName, int& width, int& height, int& channels, vec3& origin, vec3 directions[2], vec3& spacing, float& slope, float& intercept, std::vector<float>& data)
{
	float maxValue = std::numeric_limits<float>::min();
	float minValue = std::numeric_limits<float>::max();
	gdcm::ImageReader ir;
	ir.SetFileName(fileName.c_str());
	if (!ir.Read())
		return false;

	const gdcm::Image& gimage = ir.GetImage();
	const unsigned int* dimension = gimage.GetDimensions();
	width = dimension[0];
	height = dimension[1];

	const double* mOrigin = gimage.GetOrigin();
	const double* mSpacing = gimage.GetSpacing();
	origin.x = (float)mOrigin[0];
	origin.y = (float)mOrigin[1];
	origin.z = (float)mOrigin[2];
	spacing.x = (float)mSpacing[0];
	spacing.y = (float)mSpacing[1];
	spacing.z = (float)mSpacing[2];
	slope = (float)gimage.GetSlope();
	intercept = (float)gimage.GetIntercept();
	const double* mDirection = gimage.GetDirectionCosines();
	directions[0].x = (float)mDirection[0];
	directions[0].y = (float)mDirection[1];
	directions[0].z = (float)mDirection[2];
	directions[1].x = (float)mDirection[3];
	directions[1].y = (float)mDirection[4];
	directions[1].z = (float)mDirection[5];

#ifdef PRINT
//	std::cout << '\n' << width << " x " << height
//		<< ", org: " << origin.x << ", " << origin.y << ", " << origin.z
//		<< ", sp: " << spacing.x << ", " << spacing.y << ", " << spacing.z
//		<< "direction0 " << directions[0].x << ", " << directions[0].y << ", " << directions[0].z
//		<< "direction1 " << directions[1].x << ", " << directions[1].y << ", " << directions[1].z;
#endif

	switch (gimage.GetPhotometricInterpretation())
	{
	case gdcm::PhotometricInterpretation::MONOCHROME1:
	case gdcm::PhotometricInterpretation::MONOCHROME2:
		data.resize(width * height);
		channels = 1;
		switch (gimage.GetPixelFormat())
		{
		case gdcm::PixelFormat::INT16:
		{
			std::vector<short> vbuffer;
			vbuffer.resize(gimage.GetBufferLength() / sizeof(short));
			gimage.GetBuffer((char*)&vbuffer[0]);
			for (size_t i = 0; i < width * height; ++i)
			{
				float value = vbuffer[i];
				if (value > maxValue) maxValue = value;
				if (value < minValue) minValue = value;
				data[i] = value;
			}
		}
		break;
		case gdcm::PixelFormat::UINT16: // ?
		{
			std::vector<uint16_t> vbuffer;
			vbuffer.resize(gimage.GetBufferLength() / sizeof(uint16_t));
			gimage.GetBuffer((char*)&vbuffer[0]);
			for (size_t i = 0; i < width * height; ++i)
			{
				float value = vbuffer[i];
				if (value > maxValue) maxValue = value;
				if (value < minValue) minValue = value;
				data[i] = value;
			}
		}
		break;
		case gdcm::PixelFormat::INT8:
		{
			std::vector<char> vbuffer;
			vbuffer.resize(gimage.GetBufferLength() / sizeof(char));
			gimage.GetBuffer((char*)&vbuffer[0]);
			for (size_t i = 0; i < width * height; ++i)
			{
				float value = vbuffer[i];
				if (value > maxValue) maxValue = value;
				if (value < minValue) minValue = value;
				data[i] = value;
			}
		}
		break;
		case gdcm::PixelFormat::UINT8: // ?
		{
			std::vector<uint8_t> vbuffer;
			vbuffer.resize(gimage.GetBufferLength() / sizeof(uint8_t));
			gimage.GetBuffer((char*)&vbuffer[0]);
			for (size_t i = 0; i < width * height; ++i)
			{
				float value = vbuffer[i];
				if (value > maxValue) maxValue = value;
				if (value < minValue) minValue = value;
				data[i] = value;
			}
		}
		break;
		default:
			std::cout << "Unsupported pixel format: " << gimage.GetPixelFormat().GetScalarTypeAsString() << std::endl;
			return false;
		}
		break;

	case gdcm::PhotometricInterpretation::RGB:
		channels = 3;
		switch (gimage.GetPixelFormat())
		{
		case gdcm::PixelFormat::UINT16:
		{
			std::vector<uint16_t> vbuffer;
			vbuffer.resize(gimage.GetBufferLength() / sizeof(uint16_t));
			gimage.GetBuffer((char*)&vbuffer[0]);

			for (size_t i = 0; i < width * height * 3; i += 3)
			{
				float value = std::max(vbuffer[i], std::max((vbuffer[i + 1]), vbuffer[i + 2]));
				if (value > maxValue) maxValue = value;
				if (value < minValue) minValue = value;
				data[i] = value;
			}
		}
		break;
		case gdcm::PixelFormat::INT16: // ?
		{
			std::vector<short> vbuffer;
			vbuffer.resize(gimage.GetBufferLength() / sizeof(short));
			gimage.GetBuffer((char*)&vbuffer[0]);

			for (size_t i = 0; i < width * height * 3; i += 3)
			{
				float value = std::max(vbuffer[i], std::max((vbuffer[i + 1]), vbuffer[i + 2]));
				if (value > maxValue) maxValue = value;
				if (value < minValue) minValue = value;
				data[i] = value;
			}
		}
		break;

		case gdcm::PixelFormat::UINT8:
		{
			std::vector<uint8_t> vbuffer;
			vbuffer.resize(gimage.GetBufferLength() / sizeof(uint8_t));
			gimage.GetBuffer((char*)&vbuffer[0]);

			for (size_t i = 0; i < width * height * 3; i += 3)
			{
				float value = std::max(vbuffer[i], std::max((vbuffer[i + 1]), vbuffer[i + 2]));
				if (value > maxValue) maxValue = value;
				if (value < minValue) minValue = value;
				data[i] = value;
			}
		}
		break;
		case gdcm::PixelFormat::INT8:
		{
			std::vector<char> vbuffer;
			vbuffer.resize(gimage.GetBufferLength() / sizeof(char));
			gimage.GetBuffer((char*)&vbuffer[0]);

			for (size_t i = 0; i < width * height * 3; i += 3)
			{
				float value = std::max(vbuffer[i], std::max((vbuffer[i + 1]), vbuffer[i + 2]));
				if (value > maxValue) maxValue = value;
				if (value < minValue) minValue = value;
				data[i] = value;
			}
		}
		break;
		default:
			std::cout << "Unsupported pixel format: " << gimage.GetPixelFormat().GetScalarTypeAsString() << std::endl;
			return false;
		}
	}
	//bool invert = (gimage.GetPhotometricInterpretation() == gdcm::PhotometricInterpretation::MONOCHROME1);
	return true;
}

bool loadDicomDir(const std::string& fileName)
{
	return false;
}

bool saveDicomImage(const std::string& fileName, int width, int height, int channels, const std::vector<uint16_t>& data, const std::string& referenceFileName)
{
	bool useRef = fs::is_regular_file(referenceFileName);

	gdcm::ImageReader refReader;
	if (useRef)
	{
		refReader.SetFileName(referenceFileName.c_str());
		if (refReader.CanRead())
		{
			refReader.Read();
			gdcm::DataSet& refDS = refReader.GetFile().GetDataSet();
			refDS.Remove(gdcm::Tag(0x0008, 0x0018));
		}
		else {
			useRef = false;
		}
	}

	gdcm::ImageWriter iw;
	if (useRef)
	{
		iw.SetFile(refReader.GetFile());
	}
	iw.SetFileName(fileName.c_str());
	
	gdcm::Image& im = iw.GetImage();
	im.SetNumberOfDimensions(2);
	unsigned int dims[3] = {};
	dims[0] = width;
	dims[1] = height;
	im.SetDimensions(dims);
	gdcm::PixelFormat pf = gdcm::PixelFormat::UINT16;
	pf.SetSamplesPerPixel(1);
	im.SetPixelFormat(pf);
	gdcm::PhotometricInterpretation pi = gdcm::PhotometricInterpretation::MONOCHROME2;
	im.SetPhotometricInterpretation(pi);

	gdcm::DataElement pixeldata(gdcm::Tag(0x7fe0, 0x0010));
	pixeldata.SetByteValue((const char*)data.data(), (uint32_t)data.size() * sizeof(short));
	im.SetDataElement(pixeldata);
	iw.Write();

	return true;
}

bool saveDicomImage(const std::string& fileName, const std::vector<uint16_t>& data, const DicomImageDesc& desc, const std::optional<gdcm::DataSet> refDicomDS)
{
	gdcm::ImageWriter iw;
	iw.SetFileName(fileName.c_str());

	auto& outDS = iw.GetFile().GetDataSet();
	if (refDicomDS.has_value())
	{
		outDS = refDicomDS.value();

		// generate SOP Instance UID
		gdcm::UIDGenerator id;
		auto seriesID = id.Generate();

		gdcm::DataElement instanceUidDE(gdcm::Tag(0x0008, 0x0018));
		std::stringstream iuSstr;
		iuSstr << seriesID;

		if (iuSstr.str().size() % 2) { iuSstr << " "; }
		instanceUidDE.SetByteValue(iuSstr.str().c_str(), iuSstr.str().length());
		instanceUidDE.SetVR(gdcm::VR::UI);
		outDS.Replace(instanceUidDE);
	
		// instance id
		gdcm::DataElement instanceNumDE(gdcm::Tag(0x0020, 0x0013));
		std::stringstream iidSstr;
		iidSstr << desc.slice;

		if (iidSstr.str().size() % 2) { iidSstr << " "; }
		instanceNumDE.SetByteValue(iidSstr.str().c_str(), iidSstr.str().length());
		instanceNumDE.SetVR(gdcm::VR::IS);
		outDS.Replace(instanceNumDE);
	}

	gdcm::DataElement sliceThicknesDE(gdcm::Tag(0x0018, 0x0050));
	std::stringstream stSstr;
	stSstr << desc.spacing[2];
	if (stSstr.str().size() % 2) { stSstr << " "; }
	sliceThicknesDE.SetByteValue(stSstr.str().c_str(), stSstr.str().length());
	sliceThicknesDE.SetVR(gdcm::VR::DS);
	//outDS.Insert(sliceThicknesDE);
	outDS.Replace(sliceThicknesDE);


#ifdef PRINT
	printf("Slice voxel size: %f, %f, %f\n", spacing[0], spacing[1], spacing[2]);
#endif
	
	gdcm::Image& im = iw.GetImage();
	im.SetNumberOfDimensions(2);
	unsigned int dims[2] = {};
	dims[0] = desc.width;
	dims[1] = desc.height;
	im.SetDimensions(dims);
	gdcm::PixelFormat pf = gdcm::PixelFormat::UINT16;
	pf.SetSamplesPerPixel(1);
	im.SetPixelFormat(pf);
	gdcm::PhotometricInterpretation pi = gdcm::PhotometricInterpretation::MONOCHROME2;
	im.SetPhotometricInterpretation(pi);
	im.SetSlope(desc.slope);
	im.SetIntercept(desc.intercept);
	im.SetTransferSyntax(gdcm::TransferSyntax::ExplicitVRLittleEndian);
	im.SetSpacing(0, desc.spacing[0]);
	im.SetSpacing(1, desc.spacing[1]);
	im.SetSpacing(2, desc.spacing[2]);
	im.SetOrigin(0, desc.origin[0]);
	im.SetOrigin(1, desc.origin[1]);
	im.SetOrigin(2, desc.origin[2]);

	gdcm::DataElement pixeldata(gdcm::Tag(0x7fe0, 0x0010));
	pixeldata.SetByteValue((const char*)data.data(), (uint32_t)data.size() * sizeof(short));
	im.SetDataElement(pixeldata);

	iw.Write();

	return true;
}

std::optional<gdcm::DataSet> getDataSet(const std::string& fileName)
{
	gdcm::Reader reader;
	reader.SetFileName(fileName.c_str());
	if (reader.Read())
	{
		return reader.GetFile().GetDataSet();
	}

	return {};
}

}